from __future__ import annotations

import hashlib
import os
import pickle
import tempfile
import time
import zlib
from typing import Any, Dict, Optional, Type, TypeVar
from uuid import uuid4

from .cryptography import symmetric as crypto_symmetric
from .datastructures import sdict
from .pipeline import Pipe


class SessionData(sdict):
    __slots__ = ("__sid", "__hash", "__expires", "__dump")

    def __init__(self, initial=None, sid=None, expires=None):
        sdict.__init__(self, initial or ())
        object.__setattr__(self, "_SessionData__dump", pickle.dumps(sdict(self)))
        h = hashlib.md5(self._dump).hexdigest()  # noqa: S324
        object.__setattr__(self, "_SessionData__sid", sid)
        object.__setattr__(self, "_SessionData__hash", h)
        object.__setattr__(self, "_SessionData__expires", expires)

    @property
    def _sid(self):
        return self.__sid

    @property
    def _modified(self):
        dump = pickle.dumps(sdict(self))
        h = hashlib.md5(dump).hexdigest()  # noqa: S324
        if h != self.__hash:
            object.__setattr__(self, "_SessionData__dump", dump)
            return True
        return False

    @property
    def _expiration(self):
        return self.__expires

    @property
    def _dump(self):
        # note: self.__dump is updated only on _modified call
        return self.__dump

    def _expires_after(self, value):
        object.__setattr__(self, "_SessionData__expires", value)


class SessionPipe(Pipe):
    __slots__ = ["current"]

    def __init__(
        self,
        current,
        expire: int = 3600,
        secure: bool = False,
        samesite: str = "Lax",
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None,
        cookie_data: Optional[Dict[str, Any]] = None,
    ):
        self.current = current
        self.expire = expire
        self.secure = secure
        self.samesite = samesite
        self.domain = domain
        self.cookie_name = cookie_name or f"emt_session_data_{current.app.name}"
        self.cookie_data = cookie_data or {}

    def _load_session(self, wrapper):
        raise NotImplementedError

    def _new_session(self) -> SessionData:
        raise NotImplementedError

    def _pack_session(self, expiration: int):
        self.current.response.cookies[self.cookie_name] = self._session_cookie_data()
        cookie_data = self.current.response.cookies[self.cookie_name]
        cookie_data["path"] = "/"
        cookie_data["expires"] = expiration
        cookie_data["samesite"] = self.samesite
        if self.secure:
            cookie_data["secure"] = True
        if self.domain is not None:
            cookie_data["domain"] = self.domain
        for key, val in self.cookie_data.items():
            cookie_data[key] = val

    def _close_session(self):
        expiration = self.current.session._expiration or self.expire
        self._pack_session(expiration)

    def _session_cookie_data(self) -> str:
        raise NotImplementedError

    async def open_request(self):
        if self.cookie_name in self.current.request.cookies:
            self.current.session = self._load_session(self.current.request)
        if not self.current.session:
            self.current.session = self._new_session()

    async def open_ws(self):
        if self.cookie_name in self.current.websocket.cookies:
            self.current.session = self._load_session(self.current.websocket)
        if not self.current.session:
            self.current.session = self._new_session()

    async def close_request(self):
        if getattr(self.current.response, "_done", False):
            return
        self._close_session()

    def on_stream(self):
        self.current.response._done = True
        self._close_session()

    def clear(self):
        pass


class CookieSessionPipe(SessionPipe):
    def __init__(
        self,
        current,
        key,
        expire=3600,
        secure=False,
        samesite="Lax",
        domain=None,
        cookie_name=None,
        cookie_data=None,
        compression_level=0,
    ):
        super().__init__(
            current,
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data,
        )
        self.key = key
        self.compression_level = compression_level

    def _encrypt_data(self) -> str:
        data = pickle.dumps(sdict(self.current.session))
        if self.compression_level:
            data = zlib.compress(data, self.compression_level)
        return crypto_symmetric.encrypt_b64(data, self.key)

    def _decrypt_data(self, data: str) -> SessionData:
        try:
            ddata = crypto_symmetric.decrypt_b64(data, self.key)
            if self.compression_level:
                ddata = zlib.decompress(ddata)
            rv = pickle.loads(ddata)  # noqa: S301
        except Exception:
            rv = None
        return SessionData(rv, expires=self.expire)

    def _load_session(self, wrapper) -> SessionData:
        cookie_data = wrapper.cookies[self.cookie_name].value
        return self._decrypt_data(cookie_data)

    def _new_session(self) -> SessionData:
        return SessionData(expires=self.expire)

    def _session_cookie_data(self) -> str:
        return self._encrypt_data()

    def clear(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} doesn't support sessions clearing. "
            f"Change the 'key' parameter to invalidate existing ones."
        )


class BackendStoredSessionPipe(SessionPipe):
    def _new_session(self):
        return SessionData(sid=str(uuid4()))

    def _session_cookie_data(self) -> str:
        return self.current.session._sid

    def _load_session(self, wrapper) -> Optional[SessionData]:
        sid = wrapper.cookies[self.cookie_name].value
        data = self._load(sid)
        if data is not None:
            return SessionData(data, sid=sid)
        return None

    def _delete_session(self):
        pass

    def _save_session(self, expiration: int):
        pass

    def _close_session(self):
        if not self.current.session:
            self._delete_session()
            if self.current.session._modified:
                #: if we got here means we want to destroy session definitely
                if self.cookie_name in self.current.response.cookies:
                    del self.current.response.cookies[self.cookie_name]
            return
        expiration = self.current.session._expiration or self.expire
        self._save_session(expiration)
        self._pack_session(expiration)

    def _load(self, sid: str):
        return None


class FileSessionPipe(BackendStoredSessionPipe):
    _fs_transaction_suffix = ".__emt_sess"
    _fs_mode = 0o600

    def __init__(
        self,
        current,
        expire=3600,
        secure=False,
        samesite="Lax",
        domain=None,
        cookie_name=None,
        cookie_data=None,
        filename_template="emt_%s.sess",
    ):
        super().__init__(
            current,
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data,
        )
        if filename_template.endswith(self._fs_transaction_suffix):
            raise RuntimeError(f"filename templates cannot end with {self._fs_transaction_suffix}")
        self._filename_template = filename_template
        self._path = os.path.join(self.current.app.root_path, "sessions")
        #: create required paths if needed
        if not os.path.exists(self._path):
            os.mkdir(self._path)

    def _delete_session(self):
        fn = self._get_filename(self.current.session._sid)
        try:
            os.unlink(fn)
        except OSError:
            pass

    def _save_session(self, expiration):
        if self.current.session._modified:
            self._store(self.current.session, expiration)

    def _get_filename(self, sid):
        return os.path.join(self._path, self._filename_template % str(sid))

    def _load(self, sid):
        try:
            with open(self._get_filename(sid), "rb") as f:
                exp = pickle.load(f)  # noqa: S301
                val = pickle.load(f)  # noqa: S301
        except IOError:
            return None
        if exp < time.time():
            return None
        return val

    def _store(self, session, expiration):
        fn = self._get_filename(session._sid)
        now = time.time()
        exp = now + expiration
        fd, tmp = tempfile.mkstemp(suffix=self._fs_transaction_suffix, dir=self._path)
        f = os.fdopen(fd, "wb")
        try:
            pickle.dump(exp, f, 1)
            pickle.dump(sdict(session), f, pickle.HIGHEST_PROTOCOL)
        finally:
            f.close()
        try:
            os.rename(tmp, fn)
            os.chmod(fn, self._fs_mode)
        except Exception:
            pass

    def clear(self):
        for element in os.listdir(self._path):
            try:
                os.unlink(os.path.join(self._path, element))
            except Exception:
                pass


class RedisSessionPipe(BackendStoredSessionPipe):
    def __init__(
        self,
        current,
        redis,
        prefix="emtsess:",
        expire=3600,
        secure=False,
        samesite="Lax",
        domain=None,
        cookie_name=None,
        cookie_data=None,
    ):
        super().__init__(
            current,
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data,
        )
        self.redis = redis
        self.prefix = prefix

    def _delete_session(self):
        self.redis.delete(self.prefix + self.current.session._sid)

    def _save_session(self, expiration):
        if self.current.session._modified:
            self.redis.setex(self.prefix + self.current.session._sid, expiration, self.current.session._dump)
        else:
            self.redis.expire(self.prefix + self.current.session._sid, expiration)

    def _load(self, sid):
        data = self.redis.get(self.prefix + sid)
        return pickle.loads(data) if data else data  # noqa: S301

    def clear(self):
        self.redis.delete(self.prefix + "*")


TSessionPipe = TypeVar("TSessionPipe", bound=SessionPipe)


class SessionManager:
    _pipe: Optional[SessionPipe] = None

    @classmethod
    def _build_pipe(cls, handler_cls: Type[TSessionPipe], *args: Any, **kwargs: Any) -> TSessionPipe:
        raise NotImplementedError

    @classmethod
    def cookies(
        cls,
        key: str,
        expire: int = 3600,
        secure: bool = False,
        samesite: str = "Lax",
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None,
        cookie_data: Optional[Dict[str, Any]] = None,
        compression_level: int = 0,
    ) -> CookieSessionPipe:
        return cls._build_pipe(
            CookieSessionPipe,
            key,
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data,
            compression_level=compression_level,
        )

    @classmethod
    def files(
        cls,
        expire: int = 3600,
        secure: bool = False,
        samesite: str = "Lax",
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None,
        cookie_data: Optional[Dict[str, Any]] = None,
        filename_template: str = "emt_%s.sess",
    ) -> FileSessionPipe:
        return cls._build_pipe(
            FileSessionPipe,
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data,
            filename_template=filename_template,
        )

    @classmethod
    def redis(
        cls,
        redis: Any,
        prefix: str = "emtsess:",
        expire: int = 3600,
        secure: bool = False,
        samesite: str = "Lax",
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None,
        cookie_data: Optional[Dict[str, Any]] = None,
    ) -> RedisSessionPipe:
        return cls._build_pipe(
            RedisSessionPipe,
            redis,
            prefix=prefix,
            expire=expire,
            secure=secure,
            samesite=samesite,
            domain=domain,
            cookie_name=cookie_name,
            cookie_data=cookie_data,
        )

    @classmethod
    def clear(cls):
        cls._pipe.clear()



from io import BytesIO

import pytest

from emmett_core.ctx import Current
from emmett_core.http.helpers import redirect
from emmett_core.http.response import (
    HTTPAsyncIterResponse,
    HTTPBytesResponse,
    HTTPIterResponse,
    HTTPResponse,
    HTTPStringResponse,
)
from emmett_core.http.wrappers.response import Response as _Response


class Response(_Response):
    async def stream(self, target, item_wrapper=None):
        raise NotImplementedError


@pytest.fixture(scope="function")
def current():
    rv = Current()
    rv.response = Response(None)
    return rv


@pytest.fixture(scope="function")
def rsgi_proto():
    class Transport:
        def __init__(self, proto):
            self.proto = proto

        async def send_bytes(self, data):
            self.proto.data.write(data)

    class Proto:
        def __init__(self):
            self.code = 0
            self.headers = []
            self.data = BytesIO()

        def response_stream(self, code, headers):
            self.code = code
            self.headers = headers
            return Transport(self)

    return Proto()


def test_http_string_empty():
    http = HTTPStringResponse(200)

    assert http.encoded_body == b""
    assert http.status_code == 200
    assert list(http.rsgi_headers()) == [("content-type", "text/plain")]


def test_http_bytes_empty():
    http = HTTPBytesResponse(200)

    assert http.body == b""
    assert http.status_code == 200
    assert list(http.rsgi_headers()) == [("content-type", "text/plain")]


def test_http_string():
    http = HTTPStringResponse(
        200, "Hello World", headers={"x-test": "Hello Header"}, cookies={"cookie_test": "Set-Cookie: hello cookie"}
    )

    assert http.encoded_body == b"Hello World"
    assert http.status_code == 200
    assert list(http.rsgi_headers()) == [("x-test", "Hello Header"), ("set-cookie", "hello cookie")]


@pytest.mark.asyncio
async def test_http_iter(rsgi_proto):
    def iterator():
        yield b"test"

    http = HTTPIterResponse(iterator())
    await http.rsgi(rsgi_proto)
    rsgi_proto.data.seek(0)

    assert rsgi_proto.code == 200
    assert not list(http.rsgi_headers())
    assert rsgi_proto.data.read() == b"test"


@pytest.mark.asyncio
async def test_http_aiter(rsgi_proto):
    async def iterator():
        yield b"test"

    http = HTTPAsyncIterResponse(iterator())
    await http.rsgi(rsgi_proto)
    rsgi_proto.data.seek(0)

    assert rsgi_proto.code == 200
    assert not list(http.rsgi_headers())
    assert rsgi_proto.data.read() == b"test"


def test_redirect(current):
    try:
        redirect(current, "/redirect", 302)
    except HTTPResponse as http_redirect:
        assert current.response.status == 302
        assert http_redirect.status_code == 302
        assert list(http_redirect.rsgi_headers()) == [("location", "/redirect")]



from typing import Any, Dict, Iterator, Optional, Tuple

__version__: str

#: cryptography
def aes128_cfb128_decrypt(data: bytes, key: bytes, nonce: bytes) -> bytes: ...
def aes128_cfb128_encrypt(data: bytes, key: bytes, nonce: bytes) -> bytes: ...
def aes128_cfb8_decrypt(data: bytes, key: bytes, nonce: bytes) -> bytes: ...
def aes128_cfb8_encrypt(data: bytes, key: bytes, nonce: bytes) -> bytes: ...
def aes128_ctr128(data: bytes, key: bytes, nonce: bytes) -> bytes: ...
def aes256_cfb128_decrypt(data: bytes, key: bytes, nonce: bytes) -> bytes: ...
def aes256_cfb128_encrypt(data: bytes, key: bytes, nonce: bytes) -> bytes: ...
def aes256_cfb8_decrypt(data: bytes, key: bytes, nonce: bytes) -> bytes: ...
def aes256_cfb8_encrypt(data: bytes, key: bytes, nonce: bytes) -> bytes: ...
def aes256_ctr128(data: bytes, key: bytes, nonce: bytes) -> bytes: ...
def pbkdf2_sha1(data: bytes, salt: bytes, rounds: int, klen: int) -> bytes: ...
def pbkdf2_sha256(data: bytes, salt: bytes, rounds: int, klen: int) -> bytes: ...
def pbkdf2_sha384(data: bytes, salt: bytes, rounds: int, klen: int) -> bytes: ...
def pbkdf2_sha512(data: bytes, salt: bytes, rounds: int, klen: int) -> bytes: ...

#: routing
class HTTPRouter:
    def add_static_route(
        self, route, path: str, method: str, host: Optional[str] = None, scheme: Optional[str] = None
    ): ...
    def add_re_route(
        self,
        route,
        rule: str,
        rgtmap: Dict[str, str],
        method: str,
        host: Optional[str] = None,
        scheme: Optional[str] = None,
    ): ...
    def match_route_direct(self, method: str, path: str) -> Tuple[Any, Dict[str, Any]]: ...
    def match_route_scheme(self, scheme: str, method: str, path: str) -> Tuple[Any, Dict[str, Any]]: ...
    def match_route_host(self, host: str, method: str, path: str) -> Tuple[Any, Dict[str, Any]]: ...
    def match_route_all(self, host: str, scheme: str, method: str, path: str) -> Tuple[Any, Dict[str, Any]]: ...

class WSRouter:
    def add_static_route(self, route, path: str, host: Optional[str] = None, scheme: Optional[str] = None): ...
    def add_re_route(
        self, route, rule: str, rgtmap: Dict[str, str], host: Optional[str] = None, scheme: Optional[str] = None
    ): ...
    def match_route_direct(self, path: str) -> Tuple[Any, Dict[str, Any]]: ...
    def match_route_scheme(self, scheme: str, path: str) -> Tuple[Any, Dict[str, Any]]: ...
    def match_route_host(self, host: str, path: str) -> Tuple[Any, Dict[str, Any]]: ...
    def match_route_all(self, host: str, scheme: str, path: str) -> Tuple[Any, Dict[str, Any]]: ...

#: http
def get_content_type(header_value: str) -> Optional[str]: ...

#: multipart
class MultiPartReader:
    def __init__(self, content_type_header_value: str, max_part_size: int = 1024 * 1024): ...
    def parse(self, data: bytes): ...
    def contents(self) -> MultiPartContentsIter: ...

class MultiPartContentsIter:
    def __iter__(self) -> Iterator[Tuple[str, bool, Any]]: ...

class FilePartReader:
    content_type: Optional[str]
    content_length: int
    filename: Optional[str]

    def read(self, size: Optional[int] = None) -> bytes: ...
    def __iter__(self) -> Iterator[bytes]: ...

class MultiPartEncodingError(UnicodeDecodeError): ...
class MultiPartExceedingSizeError(RuntimeError): ...
class MultiPartIOError(IOError): ...
class MultiPartParsingError(ValueError): ...
class MultiPartStateError(RuntimeError): ...



import threading
from html import escape as _htmlescape


class TagStack(threading.local):
    def __init__(self):
        self.stack = []

    def __getitem__(self, key):
        return self.stack[key]

    def append(self, item):
        self.stack.append(item)

    def pop(self, idx):
        self.stack.pop(idx)

    def __bool__(self):
        return bool(self.stack)


class HtmlTag:
    __slots__ = ["name", "components", "attributes", "stack"]

    _rules = {
        "ul": ["li"],
        "ol": ["li"],
        "table": ["tr", "thead", "tbody"],
        "thead": ["tr"],
        "tbody": ["tr"],
        "tr": ["td", "th"],
        "select": ["option", "optgroup"],
        "optgroup": ["optionp"],
    }
    _self_closed = {"br", "col", "embed", "hr", "img", "input", "link", "meta"}

    def __init__(self, name, stack=None):
        self.name = name
        self.components = []
        self.attributes = {}
        self.stack = stack
        if stack:
            stack[-1].append(self)

    def __enter__(self):
        self.stack.append(self)
        return self

    def __exit__(self, type, value, traceback):
        self.stack.pop(-1)

    @staticmethod
    def wrap(component, rules):
        if rules and (not isinstance(component, HtmlTag) or component.name not in rules):
            return HtmlTag(rules[0])(component)
        return component

    def __call__(self, *components, **attributes):
        rules = self._rules.get(self.name, [])
        self.components = [self.wrap(comp, rules) for comp in components]
        self.attributes = attributes
        return self

    def append(self, component):
        self.components.append(component)

    def insert(self, i, component):
        self.components.insert(i, component)

    def remove(self, component):
        self.components.remove(component)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.components[key]
        return self.attributes.get(key)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.components.insert(key, value)
        else:
            self.attributes[key] = value

    def __iter__(self):
        for item in self.components:
            yield item

    def __str__(self):
        return self.__html__()

    def __add__(self, other):
        return cat(self, other)

    def add_class(self, name):
        c = self["_class"]
        classes = (set(c.split()) if c else set()) | set(name.split())
        self["_class"] = " ".join(classes) if classes else None
        return self

    def remove_class(self, name):
        c = self["_class"]
        classes = (set(c.split()) if c else set()) - set(name.split())
        self["_class"] = " ".join(classes) if classes else None
        return self

    @classmethod
    def _build_html_attributes_items(cls, attrs, namespace=None):
        if namespace:
            for k, v in sorted(attrs.items()):
                nk = f"{namespace}-{k}"
                if v is True:
                    yield (nk, k)
                else:
                    yield (nk, htmlescape(v))
        else:
            for k, v in filter(lambda item: item[0].startswith("_") and item[1] is not None, sorted(attrs.items())):
                nk = k[1:]
                if isinstance(v, dict):
                    for item in cls._build_html_attributes_items(v, nk):
                        yield item
                elif v is True:
                    yield (nk, nk)
                else:
                    yield (nk, htmlescape(v))

    def _build_html_attributes(self):
        return " ".join(f'{k}="{v}"' for k, v in self._build_html_attributes_items(self.attributes))

    def __html__(self):
        name = self.name
        attrs = self._build_html_attributes()
        attrs = " " + attrs if attrs else ""
        if name in self._self_closed:
            return "<%s%s />" % (name, attrs)
        components = "".join(htmlescape(v) for v in self.components)
        return "<%s%s>%s</%s>" % (name, attrs, components, name)

    def __json__(self):
        return str(self)


class TreeHtmlTag(HtmlTag):
    __slots__ = ["parent"]

    def __init__(self, name, stack=None):
        self.parent = stack[-1] if stack else None
        super().__init__(name, stack=stack)

    def __call__(self, *components, **attributes):
        super().__call__(*components, **attributes)
        for component in filter(lambda c: isinstance(c, self.__class__), self.components):
            component.parent = self
        return self

    def append(self, component):
        super().append(component)
        if isinstance(component, self.__class__) and component.parent != self:
            component.parent = self

    def insert(self, i, component):
        super().insert(i, component)
        if isinstance(component, self.__class__) and component.parent != self:
            component.parent = self

    def remove(self, component):
        super().remove(component)
        if isinstance(component, self.__class__) and component.parent is self:
            component.parent = None


class MetaHtmlTag:
    __slots__ = ["_stack"]
    _tag_cls = HtmlTag

    def __init__(self, stack):
        self._stack = stack

    def __getattr__(self, name):
        return self._tag_cls(name, stack=self._stack)

    def __getitem__(self, name):
        return self._tag_cls(name, stack=self._stack)


class cat(HtmlTag):
    __slots__ = []

    def __init__(self, *components):
        self.components = list(components)
        self.attributes = {}

    def __html__(self):
        return "".join(htmlescape(v) for v in self.components)


def _to_str(obj):
    if not isinstance(obj, str):
        return str(obj)
    return obj


def htmlescape(obj):
    if hasattr(obj, "__html__"):
        return obj.__html__()
    return _htmlescape(_to_str(obj), True).replace("'", "&#39;")



import datetime

import pytest

from emmett_core.datastructures import sdict


def route(router, path, **kwargs):
    def wrap(f):
        return router(paths=path, **kwargs)(f)

    return wrap


@pytest.fixture(scope="function")
def cfg_http_router(current, http_router):
    @route(http_router, "/test_route")
    def test_route():
        return "Test Router"

    @route(http_router, "/test_404")
    def test_404():
        current.response.status = 404
        return "Not found, dude"

    @route(http_router, "/test2/<int:a>/<str:b>")
    def test_route2(a, b):
        return "Test Router"

    @route(http_router, "/test3/<int:a>/foo(/<str:b>)?(.<str:c>)?")
    def test_route3(a, b, c):
        return "Test Router"

    @route(http_router, "/test4/<str:a>/foo(/<int:b>)?(.<str:c>)?")
    def test_route4(a, b, c):
        return "Test Router"

    @route(http_router, "/test_int/<int:a>")
    def test_route_int(a):
        return "Test Router"

    @route(http_router, "/test_float/<float:a>")
    def test_route_float(a):
        return "Test Router"

    @route(http_router, "/test_date/<date:a>")
    def test_route_date(a):
        return "Test Router"

    @route(http_router, "/test_alpha/<alpha:a>")
    def test_route_alpha(a):
        return "Test Router"

    @route(http_router, "/test_str/<str:a>")
    def test_route_str(a):
        return "Test Router"

    @route(http_router, "/test_any/<any:a>")
    def test_route_any(a):
        return "Test Router"

    @route(http_router, "/test_complex/<int:a>/<float:b>/<date:c>/<alpha:d>/<str:e>/<any:f>")
    def test_route_complex(a, b, c, d, e, f):
        return "Test Router"

    return http_router


@pytest.fixture(scope="function")
def cfg_http_router_scheme(http_router):
    @route(http_router, "/test")
    def test_route():
        return "Test Router"

    @route(http_router, "/test2/<int:a>/<str:b>", schemes="https")
    def test_route2(a, b):
        return "Test Router"

    @route(http_router, "/test3/<int:a>/foo(/<str:b>)?(.<str:c>)?", schemes="http")
    def test_route3(a, b, c):
        return "Test Router"

    return http_router


@pytest.fixture(scope="function")
def cfg_http_router_host(http_router):
    @route(http_router, "/test")
    def test_route():
        return "Test Router"

    @route(http_router, "/test2/<int:a>/<str:b>", hostname="test1.tld")
    def test_route2(a, b):
        return "Test Router"

    return http_router


@pytest.fixture(scope="function")
def cfg_http_router_scheme_host(http_router):
    @route(http_router, "/test")
    def test_route():
        return "Test Router"

    @route(http_router, "/test2/<int:a>/<str:b>", schemes="https")
    def test_route2(a, b):
        return "Test Router"

    @route(http_router, "/test3/<int:a>/foo(/<str:b>)?(.<str:c>)?", hostname="test1.tld")
    def test_route3(a, b, c):
        return "Test Router"

    return http_router


@pytest.fixture(scope="function")
def cfg_ws_router(current, ws_router):
    @route(ws_router, "/test_route")
    def test_route():
        return

    @route(ws_router, "/test2/<int:a>/<str:b>")
    def test_route2(a, b):
        return

    @route(ws_router, "/test3/<int:a>/foo(/<str:b>)?(.<str:c>)?")
    def test_route3(a, b, c):
        return

    @route(ws_router, "/test4/<str:a>/foo(/<int:b>)?(.<str:c>)?")
    def test_route4(a, b, c):
        return

    @route(ws_router, "/test_int/<int:a>")
    def test_route_int(a):
        return

    @route(ws_router, "/test_float/<float:a>")
    def test_route_float(a):
        return

    @route(ws_router, "/test_date/<date:a>")
    def test_route_date(a):
        return

    @route(ws_router, "/test_alpha/<alpha:a>")
    def test_route_alpha(a):
        return

    @route(ws_router, "/test_str/<str:a>")
    def test_route_str(a):
        return

    @route(ws_router, "/test_any/<any:a>")
    def test_route_any(a):
        return

    @route(ws_router, "/test_complex/<int:a>/<float:b>/<date:c>/<alpha:d>/<str:e>/<any:f>")
    def test_route_complex(a, b, c, d, e, f):
        return

    return ws_router


@pytest.fixture(scope="function")
def cfg_ws_router_scheme(ws_router):
    @route(ws_router, "/test")
    def test_route():
        return

    @route(ws_router, "/test2/<int:a>/<str:b>", schemes="https")
    def test_route2(a, b):
        return

    @route(ws_router, "/test3/<int:a>/foo(/<str:b>)?(.<str:c>)?", schemes="http")
    def test_route3(a, b, c):
        return

    return ws_router


@pytest.fixture(scope="function")
def cfg_ws_router_host(ws_router):
    @route(ws_router, "/test")
    def test_route():
        return

    @route(ws_router, "/test2/<int:a>/<str:b>", hostname="test1.tld")
    def test_route2(a, b):
        return

    return ws_router


@pytest.fixture(scope="function")
def cfg_ws_router_scheme_host(ws_router):
    @route(ws_router, "/test")
    def test_route():
        return

    @route(ws_router, "/test2/<int:a>/<str:b>", schemes="https")
    def test_route2(a, b):
        return

    @route(ws_router, "/test3/<int:a>/foo(/<str:b>)?(.<str:c>)?", hostname="test1.tld")
    def test_route3(a, b, c):
        return

    return ws_router


@pytest.fixture(scope="function")
def routing_ctx(
    request,
    http_ctx_builder,
    ws_ctx_builder,
    cfg_http_router,
    cfg_ws_router,
):
    return {
        "http": sdict(router=cfg_http_router, ctx=http_ctx_builder),
        "ws": sdict(router=cfg_ws_router, ctx=ws_ctx_builder),
    }[request.param]


@pytest.fixture(scope="function")
def routing_ctx_host(
    request,
    http_ctx_builder,
    ws_ctx_builder,
    cfg_http_router_host,
    cfg_ws_router_host,
):
    return {
        "http": sdict(router=cfg_http_router_host, ctx=http_ctx_builder),
        "ws": sdict(router=cfg_ws_router_host, ctx=ws_ctx_builder),
    }[request.param]


@pytest.fixture(scope="function")
def routing_ctx_scheme(
    request,
    http_ctx_builder,
    ws_ctx_builder,
    cfg_http_router_scheme,
    cfg_ws_router_scheme,
):
    return {
        "http": sdict(router=cfg_http_router_scheme, ctx=http_ctx_builder),
        "ws": sdict(router=cfg_ws_router_scheme, ctx=ws_ctx_builder),
    }[request.param]


@pytest.fixture(scope="function")
def routing_ctx_scheme_host(
    request,
    http_ctx_builder,
    ws_ctx_builder,
    cfg_http_router_scheme_host,
    cfg_ws_router_scheme_host,
):
    return {
        "http": sdict(router=cfg_http_router_scheme_host, ctx=http_ctx_builder),
        "ws": sdict(router=cfg_ws_router_scheme_host, ctx=ws_ctx_builder),
    }[request.param]


@pytest.mark.parametrize(("routing_ctx"), ["http", "ws"], indirect=True)
@pytest.mark.parametrize(
    ("path", "name"),
    [
        ("/test_route", "test_route"),
        ("/test2/1/test", "test_route2"),
        ("/test3/1/foo", "test_route3"),
        ("/test3/1/foo/bar", "test_route3"),
        ("/test3/1/foo.baz", "test_route3"),
        ("/test3/1/foo/bar.baz", "test_route3"),
        ("/test_int/1", "test_route_int"),
        ("/test_float/1.1", "test_route_float"),
        ("/test_date/2000-01-01", "test_route_date"),
        ("/test_alpha/a", "test_route_alpha"),
        ("/test_str/a1-", "test_route_str"),
        ("/test_any/a/b", "test_route_any"),
    ],
)
def test_routing_hit(routing_ctx, path, name):
    with routing_ctx.ctx(path) as ctx:
        route, _ = routing_ctx.router.match(ctx.wrapper)
        assert route.name == f"test_router.{name}"


@pytest.mark.parametrize(("routing_ctx"), ["http", "ws"], indirect=True)
@pytest.mark.parametrize(
    "path",
    [
        "/missing",
        "/test_int",
        "/test_int/a",
        "/test_int/1.1",
        "/test_int/2000-01-01",
        "/test_float",
        "/test_float/a.a",
        "/test_float/1",
        "/test_date",
        "/test_alpha",
        "/test_alpha/a1",
        "/test_alpha/a-a",
        "/test_str",
        "/test_str/a/b",
        "/test_any",
    ],
)
def test_routing_miss(routing_ctx, path):
    with routing_ctx.ctx(path) as ctx:
        route, args = routing_ctx.router.match(ctx.wrapper)
        assert not route
        assert not args


@pytest.mark.parametrize(("routing_ctx"), ["http", "ws"], indirect=True)
def test_routing_args(routing_ctx):
    with routing_ctx.ctx("/test_complex/1/1.2/2000-12-01/foo/foo1/bar/baz") as ctx:
        route, args = routing_ctx.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route_complex"
        assert args["a"] == 1
        assert round(args["b"], 1) == 1.2
        assert args["c"] == datetime.date(2000, 12, 1)
        assert args["d"] == "foo"
        assert args["e"] == "foo1"
        assert args["f"] == "bar/baz"


@pytest.mark.parametrize(("routing_ctx_scheme"), ["http", "ws"], indirect=True)
def test_routing_with_scheme(routing_ctx_scheme):
    with routing_ctx_scheme.ctx("/test") as ctx:
        route, _ = routing_ctx_scheme.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route"

    with routing_ctx_scheme.ctx("/test", scheme="https") as ctx:
        route, _ = routing_ctx_scheme.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route"

    with routing_ctx_scheme.ctx("/test2/1/test") as ctx:
        route, _ = routing_ctx_scheme.router.match(ctx.wrapper)
        assert not route

    with routing_ctx_scheme.ctx("/test2/1/test", scheme="https") as ctx:
        route, _ = routing_ctx_scheme.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route2"

    with routing_ctx_scheme.ctx("/test3/1/foo/bar.baz") as ctx:
        route, _ = routing_ctx_scheme.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route3"

    with routing_ctx_scheme.ctx("/test3/1/foo/bar.baz", scheme="https") as ctx:
        route, _ = routing_ctx_scheme.router.match(ctx.wrapper)
        assert not route


@pytest.mark.parametrize(("routing_ctx_host"), ["http", "ws"], indirect=True)
def test_routing_with_host(routing_ctx_host):
    with routing_ctx_host.ctx("/test") as ctx:
        route, _ = routing_ctx_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route"

    with routing_ctx_host.ctx("/test", host="test1.tld") as ctx:
        route, _ = routing_ctx_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route"

    with routing_ctx_host.ctx("/test", host="test2.tld") as ctx:
        route, _ = routing_ctx_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route"

    with routing_ctx_host.ctx("/test2/1/test") as ctx:
        route, _ = routing_ctx_host.router.match(ctx.wrapper)
        assert not route

    with routing_ctx_host.ctx("/test2/1/test", host="test1.tld") as ctx:
        route, _ = routing_ctx_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route2"

    with routing_ctx_host.ctx("/test2/1/test", host="test2.tld") as ctx:
        route, _ = routing_ctx_host.router.match(ctx.wrapper)
        assert not route


@pytest.mark.parametrize(("routing_ctx_scheme_host"), ["http", "ws"], indirect=True)
def test_routing_with_scheme_and_host(routing_ctx_scheme_host):
    with routing_ctx_scheme_host.ctx("/test") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route"

    with routing_ctx_scheme_host.ctx("/test", scheme="https") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route"

    with routing_ctx_scheme_host.ctx("/test", host="test1.tld") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route"

    with routing_ctx_scheme_host.ctx("/test", scheme="https", host="test2.tld") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route"

    with routing_ctx_scheme_host.ctx("/test2/1/test") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert not route

    with routing_ctx_scheme_host.ctx("/test2/1/test", scheme="https") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route2"

    with routing_ctx_scheme_host.ctx("/test2/1/test", scheme="https", host="test1.tld") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route2"

    with routing_ctx_scheme_host.ctx("/test2/1/test", scheme="https", host="test2.tld") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route2"

    with routing_ctx_scheme_host.ctx("/test3/1/foo/bar.baz") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert not route

    with routing_ctx_scheme_host.ctx("/test3/1/foo/bar.baz", host="test1.tld") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route3"

    with routing_ctx_scheme_host.ctx("/test3/1/foo/bar.baz", scheme="https", host="test1.tld") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert route.name == "test_router.test_route3"

    with routing_ctx_scheme_host.ctx("/test3/1/foo/bar.baz", host="test2.tld") as ctx:
        route, _ = routing_ctx_scheme_host.router.match(ctx.wrapper)
        assert not route



from __future__ import annotations

import asyncio
import heapq
import pickle
import threading
import time
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union, overload

from ..typing import T
from .helpers import CacheDecorator


if TYPE_CHECKING:
    from ..routing.cache import RouteCacheRule


class CacheHandler:
    def __init__(self, prefix: str = "", default_expire: int = 300):
        self._default_expire = default_expire
        self._prefix = prefix

    @staticmethod
    def _key_prefix_(method: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(method)
        def wrap(self, key: Optional[str] = None, *args, **kwargs) -> Any:
            key = self._prefix + key if key is not None else key
            return method(self, key, *args, **kwargs)

        return wrap

    @staticmethod
    def _convert_duration_(method: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(method)
        def wrap(self, key: str, value: Any, duration: Union[int, str, None] = "default") -> Any:
            if duration is None:
                duration = 60 * 60 * 24 * 365
            if duration == "default":
                duration = self._default_expire
            now = time.time()
            return method(
                self,
                key,
                value,
                now=now,
                duration=duration,
                expiration=now + duration,  # type: ignore
            )

        return wrap

    @overload
    def __call__(
        self, key: Optional[str] = None, function: None = None, duration: Union[int, str, None] = "default"
    ) -> CacheDecorator: ...

    @overload
    def __call__(
        self, key: str, function: Optional[Callable[..., T]], duration: Union[int, str, None] = "default"
    ) -> T: ...

    def __call__(
        self,
        key: Optional[str] = None,
        function: Optional[Callable[..., T]] = None,
        duration: Union[int, str, None] = "default",
    ) -> Union[CacheDecorator, T]:
        if function:
            if asyncio.iscoroutinefunction(function):
                return self.get_or_set_loop(key, function, duration)  # type: ignore
            return self.get_or_set(key, function, duration)  # type: ignore
        return CacheDecorator(self, key, duration)

    def get_or_set(self, key: str, function: Callable[[], T], duration: Union[int, str, None] = "default") -> T:
        value = self.get(key)
        if value is None:
            value = function()
            self.set(key, value, duration)
        return value

    async def get_or_set_loop(
        self, key: str, function: Callable[[], T], duration: Union[int, str, None] = "default"
    ) -> T:
        value = self.get(key)
        if value is None:
            value = await function()  # type: ignore
            self.set(key, value, duration)
        return value

    def get(self, key: str) -> Any:
        return None

    def set(self, key: str, value: Any, duration: Union[int, str, None]):
        pass

    def clear(self, key: Optional[str] = None):
        pass

    def response(
        self,
        duration: Union[int, str, None] = "default",
        query_params: bool = True,
        language: bool = True,
        hostname: bool = False,
        headers: List[str] = [],
    ) -> RouteCacheRule:
        from ..routing.cache import RouteCacheRule

        return RouteCacheRule(self, query_params, language, hostname, headers, duration)


class RamElement:
    __slots__ = ("value", "exp", "acc")

    def __init__(self, value: Any, exp: int, acc: int):
        self.value = value
        self.exp = exp
        self.acc = acc


class RamCache(CacheHandler):
    lock = threading.RLock()

    def __init__(self, prefix: str = "", threshold: int = 500, default_expire: int = 300):
        super().__init__(prefix=prefix, default_expire=default_expire)
        self.data: Dict[str, Any] = {}
        self._heap_exp: List[Tuple[int, str]] = []
        self._heap_acc: List[Tuple[float, str]] = []
        self._threshold = threshold

    def _prune(self, now):
        # remove expired items
        while self._heap_exp:
            exp, rk = self._heap_exp[0]
            if exp >= now:
                break
            self._heap_exp.remove((exp, rk))
            element = self.data.get(rk)
            if element and element.exp == exp:
                self._heap_acc.remove((self.data[rk].acc, rk))
                del self.data[rk]
        # remove threshold exceding elements
        while len(self.data) > self._threshold:
            rk = heapq.heappop(self._heap_acc)[1]
            element = self.data.get(rk)
            if element:
                self._heap_exp.remove((element.exp, rk))
                del self.data[rk]

    @CacheHandler._key_prefix_
    def get(self, key: str) -> Any:
        try:
            with self.lock:
                element = self.data[key]
                now = time.time()
                if element.exp < now:
                    return None
                self._heap_acc.remove((element.acc, key))
                element.acc = now
                heapq.heappush(self._heap_acc, (element.acc, key))
            val = element.value
        except KeyError:
            return None
        return val

    @CacheHandler._key_prefix_
    @CacheHandler._convert_duration_
    def set(self, key: str, value: Any, **kwargs):
        with self.lock:
            self._prune(kwargs["now"])
            heapq.heappush(self._heap_exp, (kwargs["expiration"], key))
            heapq.heappush(self._heap_acc, (kwargs["now"], key))
            self.data[key] = RamElement(value, kwargs["expiration"], kwargs["now"])

    @CacheHandler._key_prefix_
    def clear(self, key: Optional[str] = None):
        with self.lock:
            if key is not None:
                try:
                    rv = self.data[key]
                    self._heap_acc.remove((rv.acc, key))
                    self._heap_exp.remove((rv.exp, key))
                    del self.data[key]
                    return
                except Exception:
                    return
            self.data.clear()
            self._heap_acc = []
            self._heap_exp = []


class RedisCache(CacheHandler):
    def __init__(
        self,
        host: str = "localhost",
        port: int = 6379,
        password: Optional[str] = None,
        db: int = 0,
        prefix: str = "cache:",
        default_expire: int = 300,
        **kwargs,
    ):
        super().__init__(prefix=prefix, default_expire=default_expire)
        try:
            import redis
        except ImportError:
            raise RuntimeError("no redis module found")
        self._cache = redis.Redis(host=host, port=port, password=password, db=db, **kwargs)

    def _dump_obj(self, value: Any) -> bytes:
        if isinstance(value, int):
            return str(value).encode("ascii")
        return b"!" + pickle.dumps(value)

    def _load_obj(self, value: Any) -> Any:
        if value is None:
            return None
        if value.startswith(b"!"):
            try:
                return pickle.loads(value[1:])  # noqa: S301
            except pickle.PickleError:
                return None
        try:
            return int(value)
        except ValueError:
            return None

    @CacheHandler._key_prefix_
    def get(self, key: str) -> Any:
        return self._load_obj(self._cache.get(key))

    @CacheHandler._key_prefix_
    @CacheHandler._convert_duration_
    def set(self, key: str, value: Any, **kwargs):
        dumped = self._dump_obj(value)
        return self._cache.setex(name=key, time=kwargs["duration"], value=dumped)

    @CacheHandler._key_prefix_
    def clear(self, key: Optional[str] = None):
        if key is not None:
            if key.endswith("*"):
                keys = self._cache.delete(self._cache.keys(key))
                if keys:
                    self._cache.delete(*keys)
                return
            self._cache.delete(key)
            return
        if self._prefix:
            keys = self._cache.keys(self._prefix + "*")
            if keys:
                self._cache.delete(*keys)
            return
        self._cache.flushdb()



import os
import re
from collections import namedtuple

from ....datastructures import sdict
from ._shortcuts import to_unicode


_always_safe = b"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_.-+"
_hexdigits = "0123456789ABCDEFabcdef"
_hextobyte = {(a + b).encode(): int(a + b, 16) for a in _hexdigits for b in _hexdigits}

_scheme_re = re.compile(r"^[a-zA-Z0-9+-.]+$")

_URLTuple = namedtuple("_URLTuple", ["scheme", "netloc", "path", "query", "fragment"])


class BaseURL(_URLTuple):
    __slots__ = ()

    def replace(self, **kwargs):
        return self._replace(**kwargs)

    @property
    def host(self):
        return self._split_host()[0]

    @property
    def ascii_host(self):
        rv = self.host
        if rv is not None and isinstance(rv, str):
            try:
                rv = _encode_idna(rv)
            except UnicodeError:
                rv = rv.encode("ascii", "ignore")
        return to_unicode(rv, "ascii", "ignore")

    @property
    def port(self):
        try:
            rv = int(to_unicode(self._split_host()[1]))
            if 0 <= rv <= 65535:
                return rv
        except (ValueError, TypeError):
            pass

    @property
    def auth(self):
        return self._split_netloc()[0]

    @property
    def username(self):
        rv = self._split_auth()[0]
        if rv is not None:
            return url_unquote(rv)

    @property
    def raw_username(self):
        return self._split_auth()[0]

    @property
    def password(self):
        rv = self._split_auth()[1]
        if rv is not None:
            return url_unquote(rv)

    @property
    def raw_password(self):
        return self._split_auth()[1]

    def to_url(self):
        return url_unparse(self)

    def decode_netloc(self):
        rv = _decode_idna(self.host or "")

        if ":" in rv:
            rv = "[%s]" % rv
        port = self.port
        if port is not None:
            rv = "%s:%d" % (rv, port)
        auth = ":".join(
            filter(
                None,
                [
                    url_unquote(self.raw_username or "", errors="strict", unsafe="/:%@"),
                    url_unquote(self.raw_password or "", errors="strict", unsafe="/:%@"),
                ],
            )
        )
        if auth:
            rv = "%s@%s" % (auth, rv)
        return rv

    def to_uri_tuple(self):
        return url_parse(iri_to_uri(self).encode("ascii"))

    def to_iri_tuple(self):
        return url_parse(uri_to_iri(self))

    def get_file_location(self, pathformat=None):
        if self.scheme != "file":
            return None, None

        path = url_unquote(self.path)
        host = self.netloc or None

        if pathformat is None:
            if os.name == "nt":
                pathformat = "windows"
            else:
                pathformat = "posix"

        if pathformat == "windows":
            if path[:1] == "/" and path[1:2].isalpha() and path[2:3] in "|:":
                path = path[1:2] + ":" + path[3:]
            windows_share = path[:3] in ("\\" * 3, "/" * 3)
            import ntpath

            path = ntpath.normpath(path)
            if windows_share and host is None:
                parts = path.lstrip("\\").split("\\", 1)
                if len(parts) == 2:
                    host, path = parts
                else:
                    host = parts[0]
                    path = ""
        elif pathformat == "posix":
            import posixpath

            path = posixpath.normpath(path)
        else:
            raise TypeError("Invalid path format %s" % repr(pathformat))

        if host in ("127.0.0.1", "::1", "localhost"):
            host = None

        return host, path

    def _split_netloc(self):
        if self._at in self.netloc:
            return self.netloc.split(self._at, 1)
        return None, self.netloc

    def _split_auth(self):
        auth = self._split_netloc()[0]
        if not auth:
            return None, None
        if self._colon not in auth:
            return auth, None
        return auth.split(self._colon, 1)

    def _split_host(self):
        rv = self._split_netloc()[1]
        if not rv:
            return None, None

        if not rv.startswith(self._lbracket):
            if self._colon in rv:
                return rv.split(self._colon, 1)
            return rv, None

        idx = rv.find(self._rbracket)
        if idx < 0:
            return rv, None

        host = rv[1:idx]
        rest = rv[idx + 1 :]
        if rest.startswith(self._colon):
            return host, rest[1:]
        return host, None


class URL(BaseURL):
    __slots__ = ()
    _at = "@"
    _colon = ":"
    _lbracket = "["
    _rbracket = "]"

    def __str__(self):
        return self.to_url()

    def encode_netloc(self):
        rv = self.ascii_host or ""
        if ":" in rv:
            rv = "[%s]" % rv
        port = self.port
        if port is not None:
            rv = "%s:%d" % (rv, port)
        auth = ":".join(
            filter(
                None,
                [
                    url_quote(self.raw_username or "", "utf-8", "strict", "/:%"),
                    url_quote(self.raw_password or "", "utf-8", "strict", "/:%"),
                ],
            )
        )
        if auth:
            rv = "%s@%s" % (auth, rv)
        return rv

    def encode(self, charset="utf-8", errors="replace"):
        return BytesURL(
            self.scheme.encode("ascii"),
            self.encode_netloc(),
            self.path.encode(charset, errors),
            self.query.encode(charset, errors),
            self.fragment.encode(charset, errors),
        )


class BytesURL(BaseURL):
    __slots__ = ()
    _at = b"@"
    _colon = b":"
    _lbracket = b"["
    _rbracket = b"]"

    def __str__(self):
        return self.to_url().decode("utf-8", "replace")

    def encode_netloc(self):
        return self.netloc

    def decode(self, charset="utf-8", errors="replace"):
        return URL(
            self.scheme.decode("ascii"),
            self.decode_netloc(),
            self.path.decode(charset, errors),
            self.query.decode(charset, errors),
            self.fragment.decode(charset, errors),
        )


def url_quote(string, charset="utf-8", errors="strict", safe="/:", unsafe=""):
    if not isinstance(string, (str, bytes, bytearray)):
        string = str(string)
    if isinstance(string, str):
        string = string.encode(charset, errors)
    if isinstance(safe, str):
        safe = safe.encode(charset, errors)
    if isinstance(unsafe, str):
        unsafe = unsafe.encode(charset, errors)
    safe = frozenset(bytearray(safe) + _always_safe) - frozenset(bytearray(unsafe))
    rv = bytearray()
    for char in bytearray(string):
        if char in safe:
            rv.append(char)
        else:
            rv.extend(("%%%02X" % char).encode("ascii"))
    return bytes(rv).decode("utf8")


def _unquote_to_bytes(string, unsafe=""):
    if isinstance(string, str):
        string = string.encode("utf-8")
    if isinstance(unsafe, str):
        unsafe = unsafe.encode("utf-8")
    unsafe = frozenset(bytearray(unsafe))
    bits = iter(string.split(b"%"))
    result = bytearray(next(bits, b""))
    for item in bits:
        try:
            char = _hextobyte[item[:2]]
            if char in unsafe:
                raise KeyError()
            result.append(char)
            result.extend(item[2:])
        except KeyError:
            result.extend(b"%")
            result.extend(item)
    return bytes(result)


def url_unquote(string, charset="utf-8", errors="replace", unsafe=""):
    rv = _unquote_to_bytes(string, unsafe)
    if charset is not None:
        rv = rv.decode(charset, errors)
    return rv


def url_quote_plus(string, charset="utf-8", errors="strict", safe=""):
    return url_quote(string, charset, errors, safe + " ", "+").replace(" ", "+")


def url_unquote_plus(s, charset="utf-8", errors="replace"):
    if isinstance(s, str):
        s = s.replace("+", " ")
    else:
        s = s.replace(b"+", b" ")
    return url_unquote(s, charset, errors)


def url_parse(url, scheme=None, allow_fragments=True):
    is_text_based = isinstance(url, str)
    if scheme is None:
        scheme = ""
    netloc = query = fragment = ""
    i = url.find(":")
    if i > 0 and _scheme_re.match(to_unicode(url[:i], errors="replace")):
        rest = url[i + 1 :]
        if not rest or any(c not in "0123456789" for c in rest):
            scheme, url = url[:i].lower(), rest

    if url[:2] == "//":
        delim = len(url)
        for c in "/?#":
            wdelim = url.find(c, 2)
            if wdelim >= 0:
                delim = min(delim, wdelim)
        netloc, url = url[2:delim], url[delim:]
        if ("[" in netloc and "]" not in netloc) or ("]" in netloc and "[" not in netloc):
            raise ValueError("Invalid IPv6 URL")

    if allow_fragments and "#" in url:
        url, fragment = url.split("#", 1)
    if "?" in url:
        url, query = url.split("?", 1)

    result_type = is_text_based and URL or BytesURL
    return result_type(scheme, netloc, url, query, fragment)


def url_unparse(components):
    scheme, netloc, path, query, fragment = components
    url = ""

    if netloc or (scheme and path.startswith("/")):
        if path and path[:1] != "/":
            path = "/" + path
        url = "//" + (netloc or "") + path
    elif path:
        url += path
    if scheme:
        url = scheme + ":" + url
    if query:
        url = url + "?" + query
    if fragment:
        url = url + "#" + fragment
    return url


def _url_encode_impl(obj, charset, encode_keys, sort, key):
    iterable = sdict()
    for key, values in obj.items():
        if not isinstance(values, list):
            values = [values]
        iterable[key] = values
    if sort:
        iterable = sorted(iterable, key=key)
    for key, values in iterable.items():
        for value in values:
            if value is None:
                continue
            if not isinstance(key, bytes):
                key = str(key).encode(charset)
            if not isinstance(value, bytes):
                value = str(value).encode(charset)
            yield url_quote_plus(key) + "=" + url_quote_plus(value)


def url_encode(obj, charset="utf-8", encode_keys=False, sort=False, key=None, separator=b"&"):
    separator = separator.decode("ascii")
    return separator.join(_url_encode_impl(obj, charset, encode_keys, sort, key))


def uri_to_iri(uri, charset="utf-8", errors="replace"):
    if isinstance(uri, tuple):
        uri = url_unparse(uri)
    uri = url_parse(to_unicode(uri, charset))
    path = url_unquote(uri.path, charset, errors, "%/;?")
    query = url_unquote(uri.query, charset, errors, "%;/?:@&=+,$#")
    fragment = url_unquote(uri.fragment, charset, errors, "%;/?:@&=+,$#")
    return url_unparse((uri.scheme, uri.decode_netloc(), path, query, fragment))


def iri_to_uri(iri, charset="utf-8", errors="strict"):
    if isinstance(iri, tuple):
        iri = url_unparse(iri)
    iri = url_parse(to_unicode(iri, charset, errors))

    netloc = iri.encode_netloc()
    path = url_quote(iri.path, charset, errors, "/:~+%")
    query = url_quote(iri.query, charset, errors, "%&[]:;$*()+,!?*/=")
    fragment = url_quote(iri.fragment, charset, errors, "=%&[]:;$()+,!?*/")

    return to_unicode(url_unparse((iri.scheme, netloc, path, query, fragment)))


def url_fix(s, charset="utf-8"):
    s = to_unicode(s, charset, "replace").replace("\\", "/")

    if s.startswith("file://") and s[7:8].isalpha() and s[8:10] in (":/", "|/"):
        s = "file:///" + s[7:]

    url = url_parse(s)
    path = url_quote(url.path, charset, safe="/%+$!*'(),")
    qs = url_quote_plus(url.query, charset, safe=":&%=+$!*'(),")
    anchor = url_quote_plus(url.fragment, charset, safe=":&%=+$!*'(),")
    return to_unicode(url_unparse((url.scheme, url.encode_netloc(), path, qs, anchor)))


def _encode_idna(domain):
    if not isinstance(domain, str):
        domain.decode("ascii")
        return domain

    try:
        return domain.encode("ascii")
    except UnicodeError:
        pass

    parts = domain.split(".")
    for idx, part in enumerate(parts):
        parts[idx] = part.encode("idna")
    return b".".join(parts)


def _decode_idna(domain):
    if isinstance(domain, str):
        try:
            domain = domain.encode("ascii")
        except UnicodeError:
            return domain

    parts = domain.split(b".")
    for idx, part in enumerate(parts):
        try:
            parts[idx] = part.decode("idna")
        except UnicodeError:
            parts[idx] = part.decode("ascii", "ignore")

    return ".".join(parts)


def get_host(scope, headers):
    if "x-forwarded-host" in headers:
        rv = headers["x-forwarded-host"].split(",", 1)[0].strip()
    elif "host" in headers:
        rv = headers["host"]
    else:
        rv = scope.server.split(":")[0]
        if (scope.scheme, scope.server.split(":")[1]) not in (("https", "443"), ("http", "80")):
            rv += ":{}".format(scope.server.split(":")[1])
    return rv



from __future__ import annotations

import os
import sys
from logging import Logger
from typing import Any, Awaitable, Callable, Dict, List, Optional, Set, Type, Union

from ._internal import create_missing_app_folders, get_root_path
from .datastructures import gsdict, sdict
from .extensions import Extension, ExtensionType, Signals
from .pipeline import Pipe
from .protocols.rsgi.test_client.client import EmmettTestClient
from .routing.cache import RouteCacheRule
from .routing.router import RoutingCtx, RoutingCtxGroup
from .typing import ErrorHandlerType
from .utils import cachedprop


class Config(gsdict):
    __slots__ = ()

    def __init__(self, app: App):
        self._app = app
        super().__init__(
            hostname_default=None,
            static_version=None,
            static_version_urls=False,
            url_default_namespace=None,
            request_max_content_length=None,
            request_multipart_max_size=1024 * 1024,
            request_body_timeout=None,
            response_timeout=None,
        )
        self._handle_static = True

    def __setattr__(self, key, value):
        obj = getattr(self.__class__, key, None)
        if isinstance(obj, property):
            return obj.fset(self, value)
        return super().__setattr__(key, value)

    @property
    def handle_static(self) -> bool:
        return self._handle_static

    @handle_static.setter
    def handle_static(self, value: bool):
        self._handle_static = value
        self._app._configure_handlers()


class AppModule:
    @classmethod
    def from_app(
        cls,
        app: App,
        import_name: str,
        name: str,
        static_folder: Optional[str],
        static_path: Optional[str],
        url_prefix: Optional[str],
        hostname: Optional[str],
        cache: Optional[RouteCacheRule],
        root_path: Optional[str],
        pipeline: List[Pipe],
        opts: Dict[str, Any] = {},
    ):
        return cls(
            app,
            name,
            import_name,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            pipeline=pipeline,
            **opts,
        )

    @classmethod
    def from_module(
        cls,
        appmod: AppModule,
        import_name: str,
        name: str,
        static_folder: Optional[str],
        static_path: Optional[str],
        url_prefix: Optional[str],
        hostname: Optional[str],
        cache: Optional[RouteCacheRule],
        root_path: Optional[str],
        opts: Dict[str, Any] = {},
    ):
        if "." in name:
            raise RuntimeError("Nested app modules' names should not contains dots")
        name = appmod.name + "." + name
        if url_prefix and not url_prefix.startswith("/"):
            url_prefix = "/" + url_prefix
        module_url_prefix = (appmod.url_prefix + (url_prefix or "")) if appmod.url_prefix else url_prefix
        hostname = hostname or appmod.hostname
        cache = cache or appmod.cache
        return cls(
            appmod.app,
            name,
            import_name,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=module_url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            pipeline=appmod.pipeline,
            **opts,
        )

    @classmethod
    def from_module_group(
        cls,
        appmodgroup: AppModuleGroup,
        import_name: str,
        name: str,
        static_folder: Optional[str],
        static_path: Optional[str],
        url_prefix: Optional[str],
        hostname: Optional[str],
        cache: Optional[RouteCacheRule],
        root_path: Optional[str],
        opts: Dict[str, Any] = {},
    ) -> AppModulesGrouped:
        mods = []
        for module in appmodgroup.modules:
            mod = cls.from_module(
                module,
                import_name,
                name,
                static_folder=static_folder,
                static_path=static_path,
                url_prefix=url_prefix,
                hostname=hostname,
                cache=cache,
                root_path=root_path,
                opts=opts,
            )
            mods.append(mod)
        return AppModulesGrouped(*mods)

    def module(
        self,
        import_name: str,
        name: str,
        static_folder: Optional[str] = None,
        static_path: Optional[str] = None,
        url_prefix: Optional[str] = None,
        hostname: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        root_path: Optional[str] = None,
        module_class: Optional[Type[AppModule]] = None,
        **kwargs: Any,
    ) -> AppModule:
        module_class = module_class or self.__class__
        return module_class.from_module(
            self,
            import_name,
            name,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            opts=kwargs,
        )

    def __init__(
        self,
        app: App,
        name: str,
        import_name: str,
        static_folder: Optional[str] = None,
        static_path: Optional[str] = None,
        url_prefix: Optional[str] = None,
        hostname: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        root_path: Optional[str] = None,
        pipeline: Optional[List[Pipe]] = None,
        **kwargs: Any,
    ):
        self.app = app
        self.name = name
        self.import_name = import_name
        if root_path is None:
            root_path = get_root_path(self.import_name)
        self.root_path = root_path
        if static_path and not static_path.startswith("/"):
            static_path = os.path.join(self.root_path, static_path)
        self._static_path = (
            os.path.join(self.app.static_path, static_folder)
            if static_folder
            else (static_path or self.app.static_path)
        )
        self.url_prefix = url_prefix
        self.hostname = hostname
        self.cache = cache
        self._super_pipeline = pipeline or []
        self.pipeline = []
        self.app._register_module(self)

    @property
    def pipeline(self) -> List[Pipe]:
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipeline: List[Pipe]):
        self._pipeline = self._super_pipeline + pipeline

    def route(self, paths: Optional[Union[str, List[str]]] = None, name: Optional[str] = None, **kwargs) -> RoutingCtx:
        if name is not None and "." in name:
            raise RuntimeError("App modules' route names should not contains dots")
        name = self.name + "." + (name or "")
        pipeline = kwargs.get("pipeline", [])
        if self.pipeline:
            pipeline = self.pipeline + pipeline
        kwargs["pipeline"] = pipeline
        kwargs["cache"] = kwargs.get("cache", self.cache)
        return self.app.route(paths=paths, name=name, prefix=self.url_prefix, hostname=self.hostname, **kwargs)

    def websocket(
        self, paths: Optional[Union[str, List[str]]] = None, name: Optional[str] = None, **kwargs
    ) -> RoutingCtx:
        if name is not None and "." in name:
            raise RuntimeError("App modules' websocket names should not contains dots")
        name = self.name + "." + (name or "")
        pipeline = kwargs.get("pipeline", [])
        if self.pipeline:
            pipeline = self.pipeline + pipeline
        kwargs["pipeline"] = pipeline
        return self.app.websocket(paths=paths, name=name, prefix=self.url_prefix, hostname=self.hostname, **kwargs)


class App:
    __slots__ = [
        "__dict__",
        "_asgi_handlers",
        "_extensions_env",
        "_extensions_listeners",
        "_language_default",
        "_language_force_on_url",
        "_languages_set",
        "_languages",
        "_logger",
        "_modules",
        "_pipeline",
        "_router_http",
        "_router_ws",
        "config",
        "error_handlers",
        "ext",
        "import_name",
        "logger_name",
        "root_path",
        "static_path",
    ]

    debug = None
    config_class = Config
    modules_class = AppModule
    signals_class = Signals
    test_client_class = EmmettTestClient

    def __init__(self, import_name: str, root_path: Optional[str] = None, url_prefix: Optional[str] = None, **opts):
        self.import_name = import_name
        #: init debug var
        self.debug = os.environ.get("EMMETT_RUN_ENV") == "true"
        #: set paths for the application
        self._configure_paths(root_path, opts)
        #: init the configuration
        self.config = self.config_class(self)
        #: init languages
        self._languages: List[str] = []
        self._languages_set: Set[str] = set()
        self._language_default: Optional[str] = None
        self._language_force_on_url = False
        #: init routing
        self._pipeline: List[Pipe] = []
        self._init_routers(url_prefix)
        self._asgi_handlers = {}
        self._rsgi_handlers = {}
        self.error_handlers: Dict[int, Callable[[], Awaitable[str]]] = {}
        self._init_handlers()
        #: init logger
        self._logger = None
        self.logger_name = self.import_name
        #: init extensions
        self.ext: sdict[str, Extension] = sdict()
        self._extensions_env: sdict[str, Any] = sdict()
        self._extensions_listeners: Dict[str, List[Callable[..., Any]]] = {
            str(element): [] for element in self.signals_class
        }
        #: finalise
        self._modules: Dict[str, AppModule] = {}
        self._register_with_ctx()

    def _configure_paths(self, root_path, opts):
        if root_path is None:
            root_path = get_root_path(self.import_name)
        self.root_path = root_path
        self.static_path = os.path.join(self.root_path, "static")
        create_missing_app_folders(self)

    def _init_routers(self, url_prefix):
        raise NotImplementedError

    def _init_handlers(self):
        raise NotImplementedError

    def _configure_handlers(self):
        self._asgi_handlers["http"]._configure_methods()
        self._rsgi_handlers["http"]._configure_methods()

    def _register_with_ctx(self):
        raise NotImplementedError

    @cachedprop
    def name(self):
        if self.import_name == "__main__":
            fn = getattr(sys.modules["__main__"], "__file__", None)
            if fn is None:
                rv = "__main__"
            else:
                rv = os.path.splitext(os.path.basename(fn))[0]
        else:
            rv = self.import_name
        return rv

    @property
    def languages(self) -> List[str]:
        return self._languages

    @languages.setter
    def languages(self, value: List[str]):
        self._languages = value
        self._languages_set = set(self._languages)

    @property
    def language_default(self) -> Optional[str]:
        return self._language_default

    @language_default.setter
    def language_default(self, value: str):
        self._language_default = value

    @property
    def language_force_on_url(self) -> bool:
        return self._language_force_on_url

    @language_force_on_url.setter
    def language_force_on_url(self, value: bool):
        self._language_force_on_url = value
        self._router_http._mixin_cls._set_language_impl(self._router_http)
        self._router_ws._mixin_cls._set_language_impl(self._router_ws)
        self._configure_handlers()

    @property
    def pipeline(self) -> List[Pipe]:
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipes: List[Pipe]):
        self._pipeline = pipes
        self._router_http.pipeline = self._pipeline
        self._router_ws.pipeline = self._pipeline

    def route(
        self,
        paths: Optional[Union[str, List[str]]] = None,
        name: Optional[str] = None,
        pipeline: Optional[List[Pipe]] = None,
        schemes: Optional[Union[str, List[str]]] = None,
        hostname: Optional[str] = None,
        methods: Optional[Union[str, List[str]]] = None,
        prefix: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        output: str = "auto",
    ) -> RoutingCtx:
        if callable(paths):
            raise SyntaxError("Use @route(), not @route.")
        return self._router_http(
            paths=paths,
            name=name,
            pipeline=pipeline,
            schemes=schemes,
            hostname=hostname,
            methods=methods,
            prefix=prefix,
            cache=cache,
            output=output,
        )

    def websocket(
        self,
        paths: Optional[Union[str, List[str]]] = None,
        name: Optional[str] = None,
        pipeline: Optional[List[Pipe]] = None,
        schemes: Optional[Union[str, List[str]]] = None,
        hostname: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> RoutingCtx:
        if callable(paths):
            raise SyntaxError("Use @websocket(), not @websocket.")
        return self._router_ws(
            paths=paths, name=name, pipeline=pipeline, schemes=schemes, hostname=hostname, prefix=prefix
        )

    def on_error(self, code: int) -> Callable[[ErrorHandlerType], ErrorHandlerType]:
        def decorator(f: ErrorHandlerType) -> ErrorHandlerType:
            self.error_handlers[code] = f
            return f

        return decorator

    @property
    def log(self) -> Logger:
        if self._logger and self._logger.name == self.logger_name:
            return self._logger
        from .log import _logger_lock, create_logger

        with _logger_lock:
            if self._logger and self._logger.name == self.logger_name:
                return self._logger
            self._logger = rv = create_logger(self)
            return rv

    #: Register modules
    def _register_module(self, mod: AppModule):
        self._modules[mod.name] = mod

    #: Creates the extensions' environments and configs
    def __init_extension(self, ext):
        if ext.namespace is None:
            ext.namespace = ext.__name__
        if self._extensions_env[ext.namespace] is None:
            self._extensions_env[ext.namespace] = sdict()
        return self._extensions_env[ext.namespace], self.config[ext.namespace]

    #: Register extension listeners
    def __register_extension_listeners(self, ext):
        for signal, listener in ext._listeners_:
            self._extensions_listeners[signal].append(listener)

    #: Add an extension to application
    def use_extension(self, ext_cls: Type[ExtensionType]) -> ExtensionType:
        if not issubclass(ext_cls, Extension):
            raise RuntimeError(f"{ext_cls.__name__} is an invalid extension")
        ext_env, ext_config = self.__init_extension(ext_cls)
        ext = self.ext[ext_cls.__name__] = ext_cls(self, ext_env, ext_config)
        self.__register_extension_listeners(ext)
        ext.on_load()
        return ext

    def send_signal(self, signal: Signals, *args, **kwargs):
        for listener in self._extensions_listeners[signal]:
            listener(*args, **kwargs)

    def test_client(self, use_cookies: bool = True, **kwargs):
        return self.__class__.test_client_class(self, use_cookies=use_cookies, **kwargs)

    def __call__(self, scope, receive, send):
        return self._asgi_handlers[scope["type"]](scope, receive, send)

    def __rsgi__(self, scope, protocol):
        return self._rsgi_handlers[scope.proto](scope, protocol)

    def __rsgi_init__(self, loop):
        self.send_signal(Signals.after_loop, loop=loop)

    def module(
        self,
        import_name: str,
        name: str,
        static_folder: Optional[str] = None,
        static_path: Optional[str] = None,
        url_prefix: Optional[str] = None,
        hostname: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        root_path: Optional[str] = None,
        pipeline: Optional[List[Pipe]] = None,
        module_class: Optional[Type[AppModule]] = None,
        **kwargs: Any,
    ) -> AppModule:
        module_class = module_class or self.modules_class
        return module_class.from_app(
            self,
            import_name,
            name,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            pipeline=pipeline or [],
            opts=kwargs,
        )

    def module_group(self, *modules: AppModule) -> AppModuleGroup:
        return AppModuleGroup(*modules)


class AppModuleGroup:
    def __init__(self, *modules: AppModule):
        self.modules = modules

    def module(
        self,
        import_name: str,
        name: str,
        static_folder: Optional[str] = None,
        static_path: Optional[str] = None,
        url_prefix: Optional[str] = None,
        hostname: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        root_path: Optional[str] = None,
        module_class: Optional[Type[AppModule]] = None,
        **kwargs: Any,
    ) -> AppModulesGrouped:
        module_class = module_class or AppModule
        return module_class.from_module_group(
            self,
            import_name,
            name,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            opts=kwargs,
        )

    def route(
        self, paths: Optional[Union[str, List[str]]] = None, name: Optional[str] = None, **kwargs
    ) -> RoutingCtxGroup:
        return RoutingCtxGroup([mod.route(paths=paths, name=name, **kwargs) for mod in self.modules])

    def websocket(self, paths: Optional[Union[str, List[str]]] = None, name: Optional[str] = None, **kwargs):
        return RoutingCtxGroup([mod.websocket(paths=paths, name=name, **kwargs) for mod in self.modules])


class AppModulesGrouped(AppModuleGroup):
    @property
    def pipeline(self) -> List[List[Pipe]]:
        return [module.pipeline for module in self.modules]

    @pipeline.setter
    def pipeline(self, pipeline: List[Pipe]):
        for module in self.modules:
            module.pipeline = pipeline



import re
from typing import BinaryIO, Dict, Iterator, MutableMapping, Optional, Tuple, Union

from ..._io import loop_copyfileobj


regex_client = re.compile(r"[\w\-:]+(\.[\w\-]+)*\.?")


class ResponseHeaders(MutableMapping[str, str]):
    __slots__ = ["_data"]

    def __init__(self, data: Optional[Dict[str, str]] = None):
        self._data = data or {}

    __hash__ = None  # type: ignore

    def __getitem__(self, key: str) -> str:
        return self._data[key.lower()]

    def __setitem__(self, key: str, value: str):
        self._data[key.lower()] = value

    def __delitem__(self, key: str):
        del self._data[key.lower()]

    def __contains__(self, key: str) -> bool:  # type: ignore
        return key.lower() in self._data

    def __iter__(self) -> Iterator[str]:
        for key in self._data.keys():
            yield key

    def __len__(self) -> int:
        return len(self._data)

    def items(self) -> Iterator[Tuple[str, str]]:  # type: ignore
        for key, value in self._data.items():
            yield key, value

    def keys(self) -> Iterator[str]:  # type: ignore
        for key in self._data.keys():
            yield key

    def values(self) -> Iterator[str]:  # type: ignore
        for value in self._data.values():
            yield value

    def update(self, data: Dict[str, str]):  # type: ignore
        self._data.update(data)


class FileStorage:
    __slots__ = ["file"]

    def __init__(self, file):
        self.file = file

    def __iter__(self):
        return self.file.__iter__()

    def __getattr__(self, name):
        return getattr(self.file, name)

    @property
    def size(self):
        return self.file.content_length

    async def save(self, destination: Union[BinaryIO, str], buffer_size: int = 16384):
        close_destination = False
        if isinstance(destination, str):
            destination = open(destination, "wb")
            close_destination = True
        try:
            await loop_copyfileobj(self.file, destination, buffer_size)
        finally:
            if close_destination:
                destination.close()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.filename} ({self.content_type})>"



from binascii import unhexlify
from io import StringIO


def pkcs7_pad(data, plen: int = 16):
    data_len = len(data)
    output = StringIO()
    val = plen - (data_len % plen)
    for _ in range(val):
        output.write("%02x" % val)
    return data + unhexlify(output.getvalue())


def pkcs7_unpad(data, blks: int = 1):
    return data[: -data[-blks]]



import datetime
from abc import abstractmethod
from io import BytesIO
from typing import Any, Optional
from urllib.parse import parse_qs

from ..._emmett_core import (
    MultiPartEncodingError,
    MultiPartExceedingSizeError,
    MultiPartParsingError,
    MultiPartReader,
    MultiPartStateError,
    get_content_type,
)
from ...datastructures import sdict
from ...http.response import HTTPBytesResponse
from ...parsers import Parsers
from ...utils import cachedprop
from . import IngressWrapper
from .helpers import FileStorage


class Request(IngressWrapper):
    __slots__ = ["_now", "method"]

    method: str

    @property
    @abstractmethod
    async def body(self) -> bytes: ...

    #: allow eventual overrides
    @cachedprop
    def now(self) -> datetime.datetime:
        return self._now

    @cachedprop
    def content_type(self) -> Optional[str]:
        return get_content_type(self.headers.get("content-type", "")) or "text/plain"

    @cachedprop
    def content_length(self) -> int:
        return self.headers.get("content_length", 0, cast=int)

    _empty_body_methods = {v: v for v in ["GET", "HEAD", "OPTIONS"]}

    @cachedprop
    async def _input_params(self):
        if self._empty_body_methods.get(self.method) or not self.content_type:
            return sdict(), sdict()
        return await self._load_params()

    @cachedprop
    async def body_params(self) -> sdict[str, Any]:
        rv, _ = await self._input_params
        return rv

    @cachedprop
    async def files(self) -> sdict[str, FileStorage]:
        _, rv = await self._input_params
        return rv

    @staticmethod
    async def _load_params_missing(*args, **kwargs):
        return sdict(), sdict()

    async def _load_params_json(self, body):
        try:
            params = Parsers.get_for("json")(await body)
        except Exception:
            params = {}
        return sdict(params), sdict()

    async def _load_params_form_urlencoded(self, body):
        rv = sdict()
        data = await body
        for key, values in parse_qs(data.decode("latin-1"), keep_blank_values=True).items():
            if len(values) == 1:
                rv[key] = values[0]
                continue
            rv[key] = values
        return rv, sdict()

    @property
    def _multipart_headers(self):
        return self.headers

    @staticmethod
    def _file_param_from_field(field):
        return FileStorage(BytesIO(field.file.read()), field.filename, field.name, field.type, field.headers)

    async def _load_params_form_multipart(self, body):
        params, files = sdict(), sdict()
        try:
            parser = MultiPartReader(self.headers.get("content-type"), self.max_multipart_size)
            async for chunk in body:
                parser.parse(chunk)
            for key, is_file, field in parser.contents():
                if is_file:
                    files[key] = data = files[key] or []
                    data.append(FileStorage(field))
                else:
                    params[key] = data = params[key] or []
                    data.append(field.decode("utf8"))
        except MultiPartEncodingError:
            raise HTTPBytesResponse(400, b"Invalid encoding")
        except (MultiPartParsingError, MultiPartStateError):
            raise HTTPBytesResponse(400, b"Invalid multipart data")
        except MultiPartExceedingSizeError:
            raise HTTPBytesResponse(413, b"Request entity too large")
        for target in (params, files):
            for key, val in target.items():
                if len(val) == 1:
                    target[key] = val[0]
        return params, files

    _params_loaders = {
        "application/json": _load_params_json,
        "application/x-www-form-urlencoded": _load_params_form_urlencoded,
        "multipart/form-data": _load_params_form_multipart,
    }

    def _load_params(self):
        loader = self._params_loaders.get(self.content_type, self._load_params_missing)
        return loader(self, self.body)

    @abstractmethod
    async def push_promise(self, path: str): ...



from __future__ import annotations

import asyncio
import os
import re
import time
from collections import OrderedDict
from email.utils import formatdate
from typing import Any, Awaitable, Callable, Optional, Tuple, Union

from ...ctx import RequestContext, WSContext
from ...extensions import Signals
from ...http.response import HTTPFileResponse, HTTPResponse, HTTPStringResponse
from ...utils import cachedprop
from .helpers import RequestCancelled
from .typing import Event, EventHandler, EventLooper, Receive, Scope, Send
from .wrappers import Request, Response, Websocket


REGEX_STATIC = re.compile(r"^/static/(?P<m>__[\w\-\.]+__/)?(?P<v>_\d+\.\d+\.\d+/)?(?P<f>.*?)$")
REGEX_STATIC_LANG = re.compile(r"^/(?P<l>\w{2}/)?static/(?P<m>__[\w\-\.]__+/)?(?P<v>_\d+\.\d+\.\d+/)?(?P<f>.*?)$")


class EventHandlerWrapper:
    __slots__ = ["event", "f"]

    def __init__(self, event: str, f: EventHandler):
        self.event = event
        self.f = f

    async def __call__(
        self, handler: Handler, scope: Scope, receive: Receive, send: Send, event: Event
    ) -> Tuple[Optional[EventHandler], None]:
        task = await self.f(handler, scope, receive, send, event)
        return task, None


class MetaHandler(type):
    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        declared_events = OrderedDict()
        all_events = OrderedDict()
        events = []
        for key, value in list(attrs.items()):
            if isinstance(value, EventHandlerWrapper):
                events.append((key, value))
        declared_events.update(events)
        new_class._declared_events_ = declared_events
        for base in reversed(new_class.__mro__[1:]):
            if hasattr(base, "_declared_events_"):
                all_events.update(base._declared_events_)
        all_events.update(declared_events)
        new_class._all_events_ = all_events
        new_class._events_handlers_ = {el.event: el for el in new_class._all_events_.values()}
        return new_class


class Handler(metaclass=MetaHandler):
    __slots__ = ["app", "current"]

    def __init__(self, app, current):
        self.app = app
        self.current = current

    @classmethod
    def on_event(cls, event: str) -> Callable[[EventHandler], EventHandlerWrapper]:
        def wrap(f: EventHandler) -> EventHandlerWrapper:
            return EventHandlerWrapper(event, f)

        return wrap

    def get_event_handler(self, event_type: str) -> Union[EventHandler, EventHandlerWrapper]:
        return self._events_handlers_.get(event_type, _event_missing)

    def __call__(self, scope: Scope, receive: Receive, send: Send) -> Awaitable[None]:
        return self.handle_events(scope, receive, send)

    async def handle_events(self, scope: Scope, receive: Receive, send: Send):
        task: Optional[EventLooper] = _event_looper
        event = None
        while task:
            task, event = await task(self, scope, receive, send, event)


class LifeSpanHandler(Handler):
    __slots__ = []

    @Handler.on_event("lifespan.startup")
    async def event_startup(self, scope: Scope, receive: Receive, send: Send, event: Event) -> EventLooper:
        self.app.send_signal(Signals.after_loop, loop=asyncio.get_event_loop())
        await send({"type": "lifespan.startup.complete"})
        return _event_looper

    @Handler.on_event("lifespan.shutdown")
    async def event_shutdown(self, scope: Scope, receive: Receive, send: Send, event: Event):
        await send({"type": "lifespan.shutdown.complete"})


class RequestHandler(Handler):
    __slots__ = ["router"]

    def __init__(self, app, current):
        super().__init__(app, current)
        self._bind_router()
        self._configure_methods()

    def _bind_router(self):
        raise NotImplementedError

    def _configure_methods(self):
        raise NotImplementedError


class HTTPHandler(RequestHandler):
    __slots__ = ["pre_handler", "static_handler", "static_matcher", "__dict__"]
    wrapper_cls = Request
    response_cls = Response

    def _bind_router(self):
        self.router = self.app._router_http
        self._internal_assets_md = (str(int(time.time())), formatdate(time.time(), usegmt=True))

    def _configure_methods(self):
        self.static_matcher = (
            self._static_lang_matcher if self.app.language_force_on_url else self._static_nolang_matcher
        )
        self.static_handler = self._static_handler if self.app.config.handle_static else self.dynamic_handler
        self.pre_handler = self._prefix_handler if self.router._prefix_main else self.dynamic_handler

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        scope["emt.path"] = scope["path"] or "/"
        try:
            http = await self.pre_handler(scope, receive, send)
        except RequestCancelled:
            return
        coro = http.asgi(scope, send)
        if self.app.config.response_timeout is None:
            await coro
            return
        try:
            await asyncio.wait_for(coro, self.app.config.response_timeout)
        except asyncio.TimeoutError:
            self.app.log.warn(f"Timeout sending response: ({scope['emt.path']})")

    @cachedprop
    def error_handler(self) -> Callable[[], Awaitable[str]]:
        return self.exception_handler

    @cachedprop
    def exception_handler(self) -> Callable[[], Awaitable[str]]:
        return self.app.error_handlers.get(500, self._exception_handler)

    @staticmethod
    async def _http_response(code: int) -> HTTPResponse:
        return HTTPResponse(code)

    def _prefix_handler(self, scope: Scope, receive: Receive, send: Send) -> Awaitable[HTTPResponse]:
        path = scope["emt.path"]
        if not path.startswith(self.router._prefix_main):
            return self._http_response(404)
        scope["emt.path"] = path[self.router._prefix_main_len :] or "/"
        return self.static_handler(scope, receive, send)

    def _static_lang_matcher(self, path: str) -> Tuple[Optional[str], Optional[str]]:
        match = REGEX_STATIC_LANG.match(path)
        if match:
            lang, mname, version, file_name = match.group("l", "m", "v", "f")
            if mname:
                mod = self.app._modules.get(mname)
                spath = mod._static_path if mod else self.app.static_path
            else:
                spath = self.app.static_path
            static_file = os.path.join(spath, file_name)
            if lang:
                lang_file = os.path.join(spath, lang, file_name)
                if os.path.exists(lang_file):
                    static_file = lang_file
            return static_file, version
        return None, None

    def _static_nolang_matcher(self, path: str) -> Tuple[Optional[str], Optional[str]]:
        if path.startswith("/static/"):
            mname, version, file_name = REGEX_STATIC.match(path).group("m", "v", "f")
            if mname:
                mod = self.app._modules.get(mname[2:-3])
                static_file = os.path.join(mod._static_path, file_name) if mod else None
            elif file_name:
                static_file = os.path.join(self.app.static_path, file_name)
            else:
                static_file = None
            return static_file, version
        return None, None

    async def _static_response(self, file_path: str) -> HTTPFileResponse:
        return HTTPFileResponse(file_path)

    def _static_handler(self, scope: Scope, receive: Receive, send: Send) -> Awaitable[HTTPResponse]:
        static_file, _ = self.static_matcher(scope["emt.path"])
        if static_file:
            return self._static_response(static_file)
        return self.dynamic_handler(scope, receive, send)

    async def dynamic_handler(self, scope: Scope, receive: Receive, send: Send) -> HTTPResponse:
        request = self.__class__.wrapper_cls(
            scope,
            receive,
            send,
            max_content_length=self.app.config.request_max_content_length,
            max_multipart_size=self.app.config.request_multipart_max_size,
            body_timeout=self.app.config.request_body_timeout,
        )
        response = self.response_cls(send)
        ctx = RequestContext(self.app, request, response)
        ctx_token = self.current._init_(ctx)
        try:
            http = await self.router.dispatch(request, response)
        except HTTPResponse as http_exception:
            http = http_exception
            #: render error with handlers if in app
            error_handler = self.app.error_handlers.get(http.status_code)
            if error_handler:
                http = HTTPStringResponse(
                    http.status_code, await error_handler(), headers=response.headers, cookies=response.cookies
                )
        except RequestCancelled:
            raise
        except Exception:
            self.app.log.exception("Application exception:")
            http = HTTPStringResponse(500, await self.error_handler(), headers=response.headers)
        finally:
            self.current._close_(ctx_token)
        return http

    async def _exception_handler(self) -> str:
        self.current.response.headers._data["content-type"] = "text/plain"
        return "Internal error"


class WSHandler(RequestHandler):
    __slots__ = ["pre_handler"]
    wrapper_cls = Websocket

    def _bind_router(self):
        self.router = self.app._router_ws

    def _configure_methods(self):
        self.pre_handler = self._prefix_handler if self.router._prefix_main else self.dynamic_handler

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        scope["emt.input"] = asyncio.Queue()
        task_events = asyncio.create_task(self.handle_events(scope, receive, send))
        task_request = asyncio.create_task(self.handle_request(scope, send))
        _, pending = await asyncio.wait([task_request, task_events], return_when=asyncio.FIRST_COMPLETED)
        scope["emt._flow_cancel"] = True
        _cancel_tasks(pending)

    @Handler.on_event("websocket.connect")
    async def event_connect(self, scope: Scope, receive: Receive, send: Send, event: Event) -> EventLooper:
        return _event_looper

    @Handler.on_event("websocket.disconnect")
    async def event_disconnect(self, scope: Scope, receive: Receive, send: Send, event: Event):
        return

    @Handler.on_event("websocket.receive")
    async def event_receive(self, scope: Scope, receive: Receive, send: Send, event: Event) -> EventLooper:
        await scope["emt.input"].put(event.get("bytes") or event["text"])
        return _event_looper

    async def handle_request(self, scope: Scope, send: Send):
        scope["emt.path"] = scope["path"] or "/"
        scope["emt._ws_closed"] = False
        try:
            await self.pre_handler(scope, send)
        except HTTPResponse:
            if not scope["emt._ws_closed"]:
                await send({"type": "websocket.close", "code": 1006})
        except asyncio.CancelledError:
            if not scope.get("emt._flow_cancel", False):
                self.app.log.exception("Application exception:")
        except Exception:
            if not scope["emt._ws_closed"]:
                await send({"type": "websocket.close", "code": 1006})
            self.app.log.exception("Application exception:")

    def _prefix_handler(self, scope: Scope, send: Send) -> Awaitable[None]:
        path = scope["emt.path"]
        if not path.startswith(self.router._prefix_main):
            raise HTTPResponse(404)
        scope["emt.path"] = path[self.router._prefix_main_len :] or "/"
        return self.dynamic_handler(scope, send)

    async def dynamic_handler(self, scope: Scope, send: Send):
        ctx = WSContext(self.app, self.__class__.wrapper_cls(scope, scope["emt.input"].get, send))
        ctx_token = self.current._init_(ctx)
        try:
            await self.router.dispatch(ctx.websocket)
        finally:
            if not scope.get("emt._flow_cancel", False) and ctx.websocket._accepted:
                await send({"type": "websocket.close", "code": 1000})
                scope["emt._ws_closed"] = True
            self.current._close_(ctx_token)


async def _event_looper(
    handler: Handler, scope: Scope, receive: Receive, send: Send, event: Any = None
) -> Tuple[Union[EventHandler, EventHandlerWrapper], Event]:
    event = await receive()
    return handler.get_event_handler(event["type"]), event


async def _event_missing(handler: Handler, scope: Scope, receive: Receive, send: Send, event: Event):
    raise RuntimeError(f"Event type '{event['type']}' not recognized")


def _cancel_tasks(tasks):
    for task in tasks:
        task.cancel()



import asyncio


class Dispatcher:
    __slots__ = ["f", "flow_open", "flow_close"]

    def __init__(self, route):
        self.f = route.f
        self.flow_open = route.pipeline_flow_open
        self.flow_close = route.pipeline_flow_close

    async def _parallel_flow(self, flow):
        tasks = [asyncio.create_task(method()) for method in flow]
        await asyncio.gather(*tasks, return_exceptions=True)
        for task in tasks:
            if task.exception():
                raise task.exception()

    def dispatch(self, reqargs):
        return self.f(**reqargs)


class RequestDispatcher(Dispatcher):
    __slots__ = ["response_builder"]

    def __init__(self, route, rule, response_builder):
        super().__init__(route)
        self.response_builder = response_builder

    async def dispatch(self, reqargs, response):
        return self.response_builder(await self.f(**reqargs), response)


class RequestOpenDispatcher(RequestDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        await self._parallel_flow(self.flow_open)
        return self.response_builder(await self.f(**reqargs), response)


class RequestCloseDispatcher(RequestDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        try:
            rv = self.response_builder(await self.f(**reqargs), response)
        finally:
            await asyncio.shield(self._parallel_flow(self.flow_close))
        return rv


class RequestFlowDispatcher(RequestDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        await self._parallel_flow(self.flow_open)
        try:
            rv = self.response_builder(await self.f(**reqargs), response)
        finally:
            await asyncio.shield(self._parallel_flow(self.flow_close))
        return rv


class WSOpenDispatcher(Dispatcher):
    __slots__ = []

    async def dispatch(self, reqargs):
        await self._parallel_flow(self.flow_open)
        await self.f(**reqargs)


class WSCloseDispatcher(Dispatcher):
    __slots__ = []

    async def dispatch(self, reqargs):
        try:
            await self.f(**reqargs)
        except asyncio.CancelledError:
            await asyncio.shield(self._parallel_flow(self.flow_close))
            return
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await asyncio.shield(self._parallel_flow(self.flow_close))


class WSFlowDispatcher(Dispatcher):
    __slots__ = []

    async def dispatch(self, reqargs):
        await self._parallel_flow(self.flow_open)
        try:
            await self.f(**reqargs)
        except asyncio.CancelledError:
            await asyncio.shield(self._parallel_flow(self.flow_close))
            return
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await asyncio.shield(self._parallel_flow(self.flow_close))



from __future__ import annotations

import os
from pathlib import Path

import pytest

from emmett_core.serializers import Serializers


@pytest.fixture(scope="function")
def multipart_client(current, client):
    app = client.application
    json_dump = Serializers.get_for("json")

    @app.route("/", output="str")
    async def multipart():
        rv = {"params": {}, "files": {}}
        params = await current.request.body_params
        files = await current.request.files
        for key, val in params.items():
            accum = []
            if not isinstance(val, list):
                val = [val]
            for item in val:
                accum.append(item)
            rv["params"][key] = accum
        for key, val in files.items():
            accum = []
            if not isinstance(val, list):
                val = [val]
            for item in val:
                accum.append(
                    {
                        "filename": item.filename,
                        "size": item.size,
                        "content": item.read().decode("utf8"),
                        "content_type": item.content_type,
                    }
                )
            rv["files"][key] = accum
        return json_dump(rv)

    return client


@pytest.fixture(scope="function")
def multipart_save_client(current, client, tmpdir: Path):
    app = client.application

    @app.route("/", output="str")
    async def multipart():
        target = tmpdir / "save.txt"
        files = await current.request.files
        await files.test.save(str(target))
        return ""

    return client


@pytest.fixture(scope="function")
def multipart_stream_client(current, client, tmpdir: Path):
    app = client.application

    @app.route("/", output="str")
    async def multipart():
        target = tmpdir / "save.txt"
        files = await current.request.files
        with target.open("wb") as tf:
            for chunk in files.test:
                tf.write(chunk)
        return ""

    return client


@pytest.fixture(scope="function")
def multipart_copy_client(current, client, tmpdir: Path):
    app = client.application

    @app.route("/", output="str")
    async def multipart():
        target = tmpdir / "save.txt"
        files = await current.request.files
        with target.open("wb") as tf:
            tf.write(files.test.read())
        return ""

    return client


def test_multipart_request_data(multipart_client):
    response = multipart_client.post("/", data={"some": "data"}, content_type="multipart/form-data")
    assert response.json() == {"params": {"some": ["data"]}, "files": {}}


def test_multipart_request_files(tmpdir: Path, multipart_client):
    path = tmpdir / "test.txt"
    with path.open("wb") as file:
        file.write(b"<file content>")

    with path.open("rb") as f:
        response = multipart_client.post("/", data={"test": f})
        assert response.json() == {
            "params": {},
            "files": {
                "test": [
                    {
                        "filename": str(path),
                        "size": 14,
                        "content": "<file content>",
                        "content_type": "text/plain",
                    }
                ]
            },
        }


def test_multipart_request_files_with_content_type(tmpdir: Path, multipart_client):
    path = tmpdir / "test.txt"
    with path.open("wb") as file:
        file.write(b"<file content>")

    with path.open("rb") as f:
        response = multipart_client.post("/", data={"test": (f, "test.txt", "text/plain")})
        assert response.json() == {
            "params": {},
            "files": {
                "test": [
                    {
                        "filename": "test.txt",
                        "size": 14,
                        "content": "<file content>",
                        "content_type": "text/plain",
                    }
                ]
            },
        }


def test_multipart_request_multiple_files(tmpdir: Path, multipart_client):
    path1 = tmpdir / "test1.txt"
    with path1.open("wb") as file:
        file.write(b"<file1 content>")

    path2 = tmpdir / "test2.txt"
    with path2.open("wb") as file:
        file.write(b"<file2 content>")

    with path1.open("rb") as f1, path2.open("rb") as f2:
        response = multipart_client.post(
            "/", data={"test1": (f1, "test1.txt", "text/plain"), "test2": (f2, "test2.txt", "text/plain")}
        )
        assert response.json() == {
            "params": {},
            "files": {
                "test1": [
                    {
                        "filename": "test1.txt",
                        "size": 15,
                        "content": "<file1 content>",
                        "content_type": "text/plain",
                    }
                ],
                "test2": [
                    {
                        "filename": "test2.txt",
                        "size": 15,
                        "content": "<file2 content>",
                        "content_type": "text/plain",
                    }
                ],
            },
        }


def test_multipart_multi_items(tmpdir: Path, multipart_client):
    path1 = tmpdir / "test1.txt"
    with path1.open("wb") as file:
        file.write(b"<file1 content>")

    path2 = tmpdir / "test2.txt"
    with path2.open("wb") as file:
        file.write(b"<file2 content>")

    with path1.open("rb") as f1, path2.open("rb") as f2:
        response = multipart_client.post(
            "/", data={"test1": ["abc", (f1, "test1.txt", "text/plain"), (f2, "test2.txt", "text/plain")]}
        )
        assert response.json() == {
            "params": {"test1": ["abc"]},
            "files": {
                "test1": [
                    {
                        "filename": "test1.txt",
                        "size": 15,
                        "content": "<file1 content>",
                        "content_type": "text/plain",
                    },
                    {
                        "filename": "test2.txt",
                        "size": 15,
                        "content": "<file2 content>",
                        "content_type": "text/plain",
                    },
                ]
            },
        }


def test_multipart_request_mixed_files_and_data(multipart_client):
    response = multipart_client.post(
        "/",
        data=(
            # data
            b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
            b'Content-Disposition: form-data; name="field0"\r\n\r\n'
            b"value0\r\n"
            # file
            b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
            b'Content-Disposition: form-data; name="file"; filename="file.txt"\r\n'
            b"Content-Type: text/plain\r\n\r\n"
            b"<file content>\r\n"
            # data
            b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
            b'Content-Disposition: form-data; name="field1"\r\n\r\n'
            b"value1\r\n"
            b"--a7f7ac8d4e2e437c877bb7b8d7cc549c--\r\n"
        ),
        headers=[("content-type", "multipart/form-data; boundary=a7f7ac8d4e2e437c877bb7b8d7cc549c")],
    )
    assert response.json() == {
        "files": {
            "file": [
                {
                    "filename": "file.txt",
                    "size": 14,
                    "content": "<file content>",
                    "content_type": "text/plain",
                }
            ],
        },
        "params": {
            "field0": ["value0"],
            "field1": ["value1"],
        },
    }


def test_multipart_request_with_charset_for_filename(multipart_client):
    response = multipart_client.post(
        "/",
        data=(
            # file
            b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
            b'Content-Disposition: form-data; name="file"; filename="\xe6\x96\x87\xe6\x9b\xb8.txt"\r\n'
            b"Content-Type: text/plain\r\n\r\n"
            b"<file content>\r\n"
            b"--a7f7ac8d4e2e437c877bb7b8d7cc549c--\r\n"
        ),
        headers=[("content-type", "multipart/form-data; charset=utf-8; boundary=a7f7ac8d4e2e437c877bb7b8d7cc549c")],
    )
    assert response.json() == {
        "params": {},
        "files": {
            "file": [
                {
                    "filename": "文書.txt",
                    "size": 14,
                    "content": "<file content>",
                    "content_type": "text/plain",
                }
            ]
        },
    }


def test_multipart_request_without_charset_for_filename(multipart_client):
    response = multipart_client.post(
        "/",
        data=(
            # file
            b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
            b'Content-Disposition: form-data; name="file"; filename="\xe7\x94\xbb\xe5\x83\x8f.jpg"\r\n'
            b"Content-Type: image/jpeg\r\n\r\n"
            b"<file content>\r\n"
            b"--a7f7ac8d4e2e437c877bb7b8d7cc549c--\r\n"
        ),
        headers=[("content-type", "multipart/form-data; boundary=a7f7ac8d4e2e437c877bb7b8d7cc549c")],
    )
    assert response.json() == {
        "params": {},
        "files": {
            "file": [
                {
                    "filename": "画像.jpg",
                    "size": 14,
                    "content": "<file content>",
                    "content_type": "image/jpeg",
                }
            ],
        },
    }


def test_multipart_request_with_encoded_value(multipart_client):
    response = multipart_client.post(
        "/",
        data=(
            b"--20b303e711c4ab8c443184ac833ab00f\r\n"
            b"Content-Disposition: form-data; "
            b'name="value"\r\n\r\n'
            b"Transf\xc3\xa9rer\r\n"
            b"--20b303e711c4ab8c443184ac833ab00f--\r\n"
        ),
        headers=[("content-type", "multipart/form-data; charset=utf-8; boundary=20b303e711c4ab8c443184ac833ab00f")],
    )
    assert response.json() == {"params": {"value": ["Transférer"]}, "files": {}}


def test_missing_boundary_parameter(multipart_client):
    res = multipart_client.post(
        "/",
        data=(
            # file
            b'Content-Disposition: form-data; name="file"; filename="\xe6\x96\x87\xe6\x9b\xb8.txt"\r\n'
            b"Content-Type: text/plain\r\n\r\n"
            b"<file content>\r\n"
        ),
        headers=[("content-type", "multipart/form-data; charset=utf-8")],
    )
    assert res.status == 400
    assert res.data == "Invalid multipart data"


def test_missing_name_parameter_on_content_disposition(multipart_client):
    res = multipart_client.post(
        "/",
        data=(
            # data
            b'--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\nContent-Disposition: form-data; ="field0"\r\n\r\nvalue0\r\n'
        ),
        headers=[("content-type", "multipart/form-data; boundary=a7f7ac8d4e2e437c877bb7b8d7cc549c")],
    )
    assert res.status == 400
    assert res.data == "Invalid multipart data"


def test_multipart_max_size_exceeds_limit(multipart_client):
    boundary = "------------------------4K1ON9fZkj9uCUmqLHRbbR"
    multipart_data = (
        f"--{boundary}\r\n"
        f'Content-Disposition: form-data; name="small"\r\n\r\n'
        "small content\r\n"
        f"--{boundary}\r\n"
        f'Content-Disposition: form-data; name="large"\r\n\r\n'
        + ("x" * 1024 * 1024 + "x")  # 1MB + 1 byte of data
        + "\r\n"
        f"--{boundary}--\r\n"
    ).encode("utf-8")
    res = multipart_client.post(
        "/",
        data=multipart_data,
        headers=[("content-type", f"multipart/form-data; boundary={boundary}"), ("transfer-encoding", "chunked")],
    )
    assert res.status == 413
    assert res.data == "Request entity too large"


def test_multipart_request_file_save(tmpdir: Path, multipart_save_client):
    path = tmpdir / "test.txt"
    target = tmpdir / "save.txt"
    with path.open("wb") as file:
        file.write(b"<")
        for i in range(8192 * 128):
            file.write(f"{i}".zfill(7).encode("utf8"))
        file.write(b">")

    with path.open("rb") as f:
        response = multipart_save_client.post("/", data={"test": (f, "test.txt", "text/plain")})
        assert response.status == 200

    with path.open("rb") as f1, target.open("rb") as f2:
        assert f1.read() == f2.read()


@pytest.mark.skipif(bool(os.getenv("PGO_RUN")), reason="PGO build")
def test_multipart_request_file_stream(tmpdir: Path, multipart_stream_client):
    path = tmpdir / "test.txt"
    target = tmpdir / "save.txt"
    with path.open("wb") as file:
        file.write(b"<")
        for i in range(8192 * 128):
            file.write(f"{i}".zfill(7).encode("utf8"))
        file.write(b">")

    with path.open("rb") as f:
        response = multipart_stream_client.post("/", data={"test": (f, "test.txt", "text/plain")})
        assert response.status == 200

    with path.open("rb") as f1, target.open("rb") as f2:
        assert f1.read() == f2.read()


@pytest.mark.skipif(bool(os.getenv("PGO_RUN")), reason="PGO build")
def test_multipart_request_file_copy(tmpdir: Path, multipart_copy_client):
    path = tmpdir / "test.txt"
    target = tmpdir / "save.txt"
    with path.open("wb") as file:
        file.write(b"<")
        for i in range(8192 * 128):
            file.write(f"{i}".zfill(7).encode("utf8"))
        file.write(b">")

    with path.open("rb") as f:
        response = multipart_copy_client.post("/", data={"test": (f, "test.txt", "text/plain")})
        assert response.status == 200

    with path.open("rb") as f1, target.open("rb") as f2:
        assert f1.read() == f2.read()



import pytest

from emmett_core.app import App as _App
from emmett_core.ctx import Current
from emmett_core.protocols.rsgi.test_client.client import EmmettTestClient
from emmett_core.routing.router import HTTPRouter, WebsocketRouter


class App(_App):
    def _init_routers(self, url_prefix):
        pass

    def _init_handlers(self):
        pass

    def _register_with_ctx(self):
        pass

    def _init_with_test_env(self, current):
        self._router_http = HTTPRouter(self, current)
        self._router_ws = WebsocketRouter(self, current)
        current.app = self


@pytest.fixture(scope="function")
def current():
    return Current()


@pytest.fixture(scope="function")
def app(current):
    class TestClient(EmmettTestClient):
        _current = current

    App.test_client_class = TestClient
    rv = App(__name__)
    rv._init_with_test_env(current)
    return rv


@pytest.fixture(scope="function")
def client(app):
    return app.test_client()



from abc import abstractmethod
from typing import Any, Dict, Optional

from . import IngressWrapper


class Websocket(IngressWrapper):
    __slots__ = ["_flow_receive", "_flow_send", "receive", "send"]

    def _bind_flow(self, flow_receive, flow_send):
        self._flow_receive = flow_receive
        self._flow_send = flow_send

    @abstractmethod
    async def accept(self, headers: Optional[Dict[str, str]] = None, subprotocol: Optional[str] = None): ...

    async def _accept_and_receive(self) -> Any:
        await self.accept()
        return await self.receive()

    async def _accept_and_send(self, data: Any):
        await self.accept()
        await self.send(data)

    @abstractmethod
    async def _wrapped_receive(self) -> Any: ...

    @abstractmethod
    async def _wrapped_send(self, data: Any): ...



from typing import Optional


class MetaPipe(type):
    _pipeline_methods_ = {
        "open",
        "open_request",
        "open_ws",
        "close",
        "close_request",
        "close_ws",
        "pipe",
        "pipe_request",
        "pipe_ws",
        "on_pipe_success",
        "on_pipe_failure",
        "on_receive",
        "on_send",
        "on_stream",
    }

    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        if not bases:
            return new_class
        declared_methods = cls._pipeline_methods_ & set(attrs.keys())
        new_class._pipeline_declared_methods_ = declared_methods
        all_methods = set()
        for base in reversed(new_class.__mro__[:-2]):
            if hasattr(base, "_pipeline_declared_methods_"):
                all_methods = all_methods | base._pipeline_declared_methods_
        all_methods = all_methods | declared_methods
        new_class._pipeline_all_methods_ = all_methods
        new_class._is_flow_request_responsible = bool(
            all_methods & {"pipe", "pipe_request", "on_pipe_success", "on_pipe_failure"}
        )
        new_class._is_flow_ws_responsible = bool(
            all_methods & {"pipe", "pipe_ws", "on_pipe_success", "on_pipe_failure"}
        )
        if all_methods.issuperset({"pipe", "pipe_request"}):
            raise RuntimeError(f"{name} has double pipe methods. Use `pipe` or `pipe_request`, not both.")
        if all_methods.issuperset({"pipe", "pipe_ws"}):
            raise RuntimeError(f"{name} has double pipe methods. Use `pipe` or `pipe_ws`, not both.")
        return new_class


class Pipe(metaclass=MetaPipe):
    output: Optional[str] = None

    async def open(self):
        pass

    async def open_request(self):
        pass

    async def open_ws(self):
        pass

    async def close(self):
        pass

    async def close_request(self):
        pass

    async def close_ws(self):
        pass

    async def pipe(self, next_pipe, **kwargs):
        return await next_pipe(**kwargs)

    async def pipe_request(self, next_pipe, **kwargs):
        return await next_pipe(**kwargs)

    async def pipe_ws(self, next_pipe, **kwargs):
        return await next_pipe(**kwargs)

    async def on_pipe_success(self):
        pass

    async def on_pipe_failure(self):
        pass

    def on_receive(self, data):
        return data

    def on_send(self, data):
        return data

    def on_stream(self):
        pass



import os

import pytest

from emmett_core._emmett_core import MultiPartReader


@pytest.mark.skipif(bool(os.getenv("PGO_RUN")), reason="PGO build")
def test_multipart_mixed_segmented():
    data = (
        # data
        b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
        b'Content-Disposition: form-data; name="field0"\r\n\r\n'
        b"value0\r\n"
        # file
        b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
        b'Content-Disposition: form-data; name="file"; filename="file.txt"\r\n'
        b"Content-Type: text/plain\r\n\r\n"
        b"<file content>\r\n"
        # data
        b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
        b'Content-Disposition: form-data; name="field1"\r\n\r\n'
        b"value1\r\n"
        b"--a7f7ac8d4e2e437c877bb7b8d7cc549c--\r\n"
    )

    parser = MultiPartReader("multipart/form-data; boundary=a7f7ac8d4e2e437c877bb7b8d7cc549c")
    parser.parse(data[:37])

    idx = 37
    while True:
        segment = data[idx : idx + 1]
        if not segment:
            break
        parser.parse(segment)
        idx += 1
    parsed = list(parser.contents())
    assert (parsed[0][0], parsed[0][2]) == ("field0", b"value0")
    assert (parsed[2][0], parsed[2][2]) == ("field1", b"value1")
    assert (parsed[1][0], parsed[1][2].filename, parsed[1][2].read()) == ("file", "file.txt", b"<file content>")


def test_multipart_mixed_chunked():
    data = (
        # data
        b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
        b'Content-Disposition: form-data; name="field0"\r\n\r\n'
        b"value0\r\n"
        # file
        b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
        b'Content-Disposition: form-data; name="file"; filename="file.txt"\r\n'
        b"Content-Type: text/plain\r\n\r\n"
        b"<file content>\r\n"
        # data
        b"--a7f7ac8d4e2e437c877bb7b8d7cc549c\r\n"
        b'Content-Disposition: form-data; name="field1"\r\n\r\n'
        b"value1\r\n"
        b"--a7f7ac8d4e2e437c877bb7b8d7cc549c--\r\n"
    )

    step = 97

    parser = MultiPartReader("multipart/form-data; boundary=a7f7ac8d4e2e437c877bb7b8d7cc549c")
    parser.parse(data[:step])

    idx = 1
    while True:
        segment = data[idx * step : (idx + 1) * step]
        if not segment:
            break
        parser.parse(segment)
        idx += 1
    parsed = list(parser.contents())
    assert (parsed[0][0], parsed[0][2]) == ("field0", b"value0")
    assert (parsed[2][0], parsed[2][2]) == ("field1", b"value1")
    assert (parsed[1][0], parsed[1][2].filename, parsed[1][2].read()) == ("file", "file.txt", b"<file content>")



from __future__ import annotations

import asyncio
from typing import Any, Callable, Generic, Optional, Union, overload

from .typing import T


class _cached_prop(Generic[T]):
    def __init__(self, fget: Callable[..., T], name: str, doc: Optional[str] = None):
        self.fget = fget
        self.__doc__ = doc
        self.__name__ = name

    def __get__(self, obj: Optional[object], cls: Any) -> T:
        raise NotImplementedError


def cachedprop(fget: Callable[..., T], doc: Optional[str] = None, name: Optional[str] = None) -> _cached_prop[T]:
    doc = doc or fget.__doc__
    name = name or fget.__name__
    if asyncio.iscoroutinefunction(fget):
        return _cached_prop_loop[T](fget, name, doc)
    return _cached_prop_sync[T](fget, name, doc)


class _cached_prop_sync(_cached_prop[T]):
    @overload
    def __get__(self, obj: None, cls: Any) -> _cached_prop_sync: ...

    @overload
    def __get__(self, obj: object, cls: Any) -> T: ...

    def __get__(self, obj: Optional[object], cls: Any) -> Union[_cached_prop_sync, T]:
        if obj is None:
            return self
        obj.__dict__[self.__name__] = rv = self.fget(obj)
        return rv


class _cached_awaitable_coro(Generic[T]):
    slots = ["coro_f", "obj", "_result", "_awaitable"]

    def __init__(self, coro_f: Callable[..., T], obj: object):
        self.coro_f = coro_f
        self.obj = obj
        self._awaitable = self.__fetcher

    async def __fetcher(self) -> T:
        self._result = rv = await self.coro_f(self.obj)  # type: ignore
        self._awaitable = self.__cached
        return rv

    async def __cached(self) -> T:
        return self._result

    def __await__(self):
        return self._awaitable().__await__()


class _cached_prop_loop(_cached_prop[T]):
    @overload
    def __get__(self, obj: None, cls: Any) -> _cached_prop_loop: ...

    @overload
    def __get__(self, obj: object, cls: Any) -> T: ...

    def __get__(self, obj: Optional[object], cls: Any) -> Union[_cached_prop_loop, T]:
        if obj is None:
            return self
        obj.__dict__[self.__name__] = rv = _cached_awaitable_coro[T](self.fget, obj)
        return rv  # type: ignore



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, List, Optional, Union, overload

from ..typing import T
from .handlers import RamCache
from .helpers import CacheDecorator


if TYPE_CHECKING:
    from ..routing.cache import RouteCacheRule


class Cache:
    def __init__(self, **kwargs):
        #: load handlers
        handlers = []
        for key, val in kwargs.items():
            if key == "default":
                continue
            handlers.append((key, val))
        if not handlers:
            handlers.append(("ram", RamCache()))
        #: set handlers
        for name, handler in handlers:
            setattr(self, name, handler)
        _default_handler_name = kwargs.get("default", handlers[0][0])
        self._default_handler = getattr(self, _default_handler_name)

    @overload
    def __call__(
        self, key: Optional[str] = None, function: None = None, duration: Union[int, str, None] = "default"
    ) -> CacheDecorator: ...

    @overload
    def __call__(
        self, key: str, function: Optional[Callable[..., T]], duration: Union[int, str, None] = "default"
    ) -> T: ...

    def __call__(
        self,
        key: Optional[str] = None,
        function: Optional[Callable[..., T]] = None,
        duration: Union[int, str, None] = "default",
    ) -> Union[CacheDecorator, T]:
        return self._default_handler(key, function, duration)

    def get(self, key: str) -> Any:
        return self._default_handler.get(key)

    def set(self, key: str, value: Any, duration: Union[int, str, None] = "default"):
        self._default_handler.set(key, value, duration)

    def get_or_set(self, key: str, function: Callable[..., T], duration: Union[int, str, None] = "default") -> T:
        return self._default_handler.get_or_set(key, function, duration)

    def clear(self, key: Optional[str] = None):
        self._default_handler.clear(key)

    def response(
        self,
        duration: Union[int, str, None] = "default",
        query_params: bool = True,
        language: bool = True,
        hostname: bool = False,
        headers: List[str] = [],
    ) -> RouteCacheRule:
        return self._default_handler.response(duration, query_params, language, hostname, headers)



from typing import Any, Dict

from ..http.wrappers.response import ServerSentEvent
from ..routing.router import RouterMixin
from .extras import JSONPipe, RequirePipe, SSEPipe, StreamPipe


class PipeBuilder(object):
    def build_pipe(self):
        raise NotImplementedError

    def __call__(self, func):
        obj = RouterMixin.exposing()
        obj.pipeline.append(self.build_pipe())
        return func


class ServicePipeBuilder:
    _pipe_cls = {"json": JSONPipe}

    def __call__(self, procedure):
        pipe_cls = self.__class__._pipe_cls.get(procedure)
        if not pipe_cls:
            raise RuntimeError(f"Unknwon service: {procedure}")
        return pipe_cls()


class requires(PipeBuilder):
    _pipe_cls = RequirePipe

    def __init__(self, condition=None, otherwise=None):
        if condition is None or otherwise is None:
            raise SyntaxError("requires usage: @requires(condition, otherwise)")
        if not callable(otherwise) and not isinstance(otherwise, str):
            raise SyntaxError("requires 'otherwise' param must be string or callable")
        self.condition = condition
        self.otherwise = otherwise

    def build_pipe(self):
        return self.__class__._pipe_cls(self.condition, self.otherwise)


class service(PipeBuilder):
    _inner_builder = ServicePipeBuilder()

    def __init__(self, procedure):
        self.procedure = procedure

    @classmethod
    def json(cls, f):
        return cls("json")(f)

    def build_pipe(self):
        return self._inner_builder(self.procedure)


class stream(PipeBuilder):
    _pipe_cls = StreamPipe

    def __init__(
        self,
        status: int = 200,
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {},
    ):
        self.response_status = status
        self.response_headers = headers
        self.response_cookies = cookies

    def build_pipe(self):
        return self.__class__._pipe_cls(
            status=self.response_status,
            headers=self.response_headers,
            cookies=self.response_cookies,
        )


class sse(stream):
    _pipe_cls = SSEPipe
    Event = ServerSentEvent



import hmac
import os
from base64 import urlsafe_b64decode, urlsafe_b64encode
from binascii import hexlify, unhexlify
from typing import Tuple, Union

from .. import _emmett_core


def encrypt(data: Union[bytes, str], key: str) -> Tuple[bytes, bytes, bytes]:
    if isinstance(data, str):
        data = data.encode("utf8")
    klen, ivlen, salt = 256 // 8, 128 // 8, os.urandom(32)
    dkey = _emmett_core.pbkdf2_sha256(key.encode("utf8"), salt, 1000, klen * 2 + ivlen)
    k1 = dkey[:klen]
    k2 = dkey[klen : (klen * 2)]
    nonce = dkey[(klen * 2) : (klen * 2) + ivlen]
    cipher = _emmett_core.aes256_ctr128(data, k1, nonce)
    signature = hmac.digest(k2, cipher, "sha256")
    return cipher, salt, signature


def decrypt(data: bytes, salt: bytes, signature: bytes, key: str) -> bytes:
    klen, ivlen = 256 // 8, 128 // 8
    dkey = _emmett_core.pbkdf2_sha256(key.encode("utf8"), salt, 1000, klen * 2 + ivlen)
    k1 = dkey[:klen]
    k2 = dkey[klen : (klen * 2)]
    nonce = dkey[(klen * 2) : (klen * 2) + ivlen]
    if not hmac.compare_digest(signature, hmac.digest(k2, data, "sha256")):
        raise ValueError("Signature verification failed")
    return _emmett_core.aes256_ctr128(data, k1, nonce)


def encrypt_hex(data: Union[bytes, str], key: str, jchar: str = ":") -> str:
    cipher, salt, signature = encrypt(data, key)
    return jchar.join(hexlify(v).decode("utf8") for v in [salt, signature, cipher])


def decrypt_hex(data: Union[bytes, str], key: str, jchar: str = ":") -> bytes:
    try:
        salt, signature, cipher = data.split(jchar)
    except ValueError:
        raise ValueError("Invalid data input")
    return decrypt(unhexlify(cipher), unhexlify(salt), unhexlify(signature), key)


def encrypt_b64(data: Union[bytes, str], key: str, jchar: str = ":") -> str:
    cipher, salt, signature = encrypt(data, key)
    return jchar.join(urlsafe_b64encode(v).decode("utf8") for v in [salt, signature, cipher])


def decrypt_b64(data: Union[bytes, str], key: str, jchar: str = ":") -> bytes:
    try:
        salt, signature, cipher = data.split(jchar)
    except ValueError:
        raise ValueError("Invalid data input")
    return decrypt(urlsafe_b64decode(cipher), urlsafe_b64decode(salt), urlsafe_b64decode(signature), key)



from .response import HTTPRedirectResponse, HTTPStringResponse


def abort(current, code: int, body: str = ""):
    response = current.response
    response.status = code
    raise HTTPStringResponse(code, body=body, cookies=response.cookies)


def redirect(current, location: str, status_code: int = 303):
    response = current.response
    response.status = status_code
    raise HTTPRedirectResponse(status_code, location, response.cookies)



from __future__ import annotations

import asyncio
from functools import wraps
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Optional, Union

from .hash import CacheHashMixin


if TYPE_CHECKING:
    from .handlers import CacheHandler


class CacheDecorator(CacheHashMixin):
    def __init__(self, handler: CacheHandler, key: Optional[str], duration: Union[int, str, None] = "default"):
        super().__init__()
        self._cache = handler
        self.key = key
        self.duration = duration
        self.add_strategy("args")
        self.add_strategy("kwargs", self.dict_strategy)

    def _key_from_wrapped(self, f: Callable[..., Any]) -> str:
        return f.__module__ + "." + f.__name__

    def _wrap_sync(self, f: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(f)
        def wrap(*args, **kwargs) -> Any:
            if not args and not kwargs:
                key = self.key or self._key_from_wrapped(f)
            else:
                key = self._build_ctx_key(args=args, kwargs=kwargs)
            return self._cache.get_or_set(key, lambda: f(*args, **kwargs), self.duration)

        return wrap

    def _wrap_loop(self, f: Callable[..., Awaitable[Any]]) -> Callable[..., Awaitable[Any]]:
        @wraps(f)
        async def wrap(*args, **kwargs) -> Any:
            if not args and not kwargs:
                key = self.key or self._key_from_wrapped(f)
            else:
                key = self._build_ctx_key(args=args, kwargs=kwargs)
            return await self._cache.get_or_set_loop(key, lambda: f(*args, **kwargs), self.duration)

        return wrap

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        rv = self._wrap_loop(f) if asyncio.iscoroutinefunction(f) else self._wrap_sync(f)
        if not self.key:
            self.key = f.__module__ + "." + f.__name__
        return rv



import asyncio
from functools import partial
from shutil import copyfileobj


class LoopFileWrapper:
    __slots__ = ("_file", "_loop")

    def __init__(self, f, loop=None):
        self._file = f
        self._loop = loop or asyncio.get_running_loop()

    async def read(self, *args, **kwargs):
        return await self._loop.run_in_executor(None, partial(self._file.read, *args, **kwargs))

    async def write(self, *args, **kwargs):
        return await self._loop.run_in_executor(None, partial(self._file.write, *args, **kwargs))

    async def close(self, *args, **kwargs):
        return await self._loop.run_in_executor(None, partial(self._file.close, *args, **kwargs))

    def __getattr__(self, name):
        return getattr(self._file, name)


class LoopFileCtxWrapper:
    __slots__ = ("_coro", "_obj")

    def __init__(self, coro):
        self._coro = coro
        self._obj = None

    def __await__(self):
        return self._coro.__await__()

    async def __aenter__(self):
        self._obj = await self._coro
        return self._obj

    async def __aexit__(self, exc_type, exc, tb):
        await self._obj.close()
        self._obj = None


async def _loop_open_file(loop, *args, **kwargs):
    f = await loop.run_in_executor(None, partial(open, *args, **kwargs))
    return LoopFileWrapper(f, loop)


def loop_open_file(*args, **kwargs):
    return LoopFileCtxWrapper(_loop_open_file(asyncio.get_running_loop(), *args, **kwargs))


async def loop_copyfileobj(fsrc, fdst, length=None):
    return await asyncio.get_running_loop().run_in_executor(None, partial(copyfileobj, fsrc, fdst, length))



from datetime import datetime
from typing import Any, Callable, Dict, Iterator, List, Mapping, Optional, Tuple, Union
from urllib.parse import parse_qs

from ...datastructures import sdict
from ...http.response import HTTPBytesResponse, HTTPResponse
from ...http.wrappers.helpers import regex_client
from ...http.wrappers.request import Request as _Request
from ...http.wrappers.response import Response as _Response
from ...http.wrappers.websocket import Websocket as _Websocket
from ...utils import cachedprop
from .helpers import BodyWrapper, ResponseStream
from .typing import Receive, Scope, Send


_push_headers = {"accept", "accept-encoding", "accept-language", "cache-control", "user-agent"}


class Headers(Mapping[str, str]):
    __slots__ = ["_data"]

    def __init__(self, scope: Dict[str, Any]):
        self._data: Dict[bytes, bytes] = dict(scope["headers"])

    __hash__ = None  # type: ignore

    def __getitem__(self, key: str) -> str:
        return self._data[key.lower().encode("latin-1")].decode("latin-1")

    def __contains__(self, key: str) -> bool:  # type: ignore
        return key.lower().encode("latin-1") in self._data

    def __iter__(self) -> Iterator[str]:
        for key in self._data.keys():
            yield key.decode("latin-1")

    def __len__(self) -> int:
        return len(self._data)

    def get(self, key: str, default: Optional[Any] = None, cast: Optional[Callable[[Any], Any]] = None) -> Any:
        rv = self._data.get(key.lower().encode("latin-1"))
        rv = rv.decode() if rv is not None else default  # type: ignore
        if cast is None:
            return rv
        try:
            return cast(rv)
        except ValueError:
            return default

    def items(self) -> Iterator[Tuple[str, str]]:  # type: ignore
        for key, value in self._data.items():
            yield key.decode("latin-1"), value.decode("latin-1")

    def keys(self) -> Iterator[str]:  # type: ignore
        for key in self._data.keys():
            yield key.decode("latin-1")

    def values(self) -> Iterator[str]:  # type: ignore
        for value in self._data.values():
            yield value.decode("latin-1")


class ASGIIngressMixin:
    def __init__(self, scope: Scope, receive: Receive, send: Send):
        self._scope = scope
        self._receive = receive
        self._send = send
        self._scheme = scope["scheme"]
        self.path = scope["emt.path"]

    @cachedprop
    def headers(self) -> Headers:
        return Headers(self._scope)

    @cachedprop
    def query_params(self) -> sdict[str, Union[str, List[str]]]:
        rv: sdict[str, Any] = sdict()
        for key, values in parse_qs(self._scope["query_string"].decode("latin-1"), keep_blank_values=True).items():
            if len(values) == 1:
                rv[key] = values[0]
                continue
            rv[key] = values
        return rv

    @cachedprop
    def client(self) -> str:
        g = regex_client.search(self.headers.get("x-forwarded-for", ""))
        client = (g.group() or "").split(",")[0] if g else (self._scope["client"][0] if self._scope["client"] else None)
        if client in (None, "", "unknown", "localhost"):
            client = "::1" if self.host.startswith("[") else "127.0.0.1"
        return client  # type: ignore


class Request(ASGIIngressMixin, _Request):
    __slots__ = ["_scope", "_receive", "_send", "_scheme"]

    def __init__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        max_content_length: Optional[int] = None,
        max_multipart_size: Optional[int] = None,
        body_timeout: Optional[int] = None,
    ):
        super().__init__(scope, receive, send)
        self.max_content_length = max_content_length
        self.max_multipart_size = max_multipart_size
        self.body_timeout = body_timeout
        self._now = datetime.utcnow()
        self.method = scope["method"]

    @cachedprop
    def scheme(self):
        return self._scheme

    @cachedprop
    def body(self) -> BodyWrapper:
        if self.max_content_length and self.content_length > self.max_content_length:
            raise HTTPBytesResponse(413, b"Request entity too large")
        return BodyWrapper(self._receive, self.body_timeout)

    async def push_promise(self, path: str):
        if "http.response.push" not in self._scope.get("extensions", {}):
            return
        await self._send(
            {
                "type": "http.response.push",
                "path": path,
                "headers": [
                    (key.encode("latin-1"), self.headers[key].encode("latin-1"))
                    for key in _push_headers & set(self.headers.keys())
                ],
            }
        )


class Response(_Response):
    async def stream(self, target, item_wrapper=None) -> HTTPResponse:
        return await ResponseStream(self, target, item_wrapper=item_wrapper)


class Websocket(ASGIIngressMixin, _Websocket):
    __slots__ = ["_scope", "_receive", "_send", "_scheme", "_accepted"]

    def __init__(self, scope: Scope, receive: Receive, send: Send):
        super().__init__(scope, receive, send)
        self._accepted = False
        self._flow_receive = None
        self._flow_send = None
        self.receive = self._accept_and_receive
        self.send = self._accept_and_send

    @cachedprop
    def scheme(self):
        return {"ws": "http", "wss": "https"}[self._scheme]

    @property
    def _asgi_spec_version(self) -> int:
        return int("".join(self._scope.get("asgi", {}).get("spec_version", "2.0").split(".")))

    def _encode_headers(self, headers: Dict[str, str]) -> List[Tuple[bytes, bytes]]:
        return [(key.encode("utf-8"), val.encode("utf-8")) for key, val in headers.items()]

    async def accept(self, headers: Optional[Dict[str, str]] = None, subprotocol: Optional[str] = None):
        if self._accepted:
            return
        message: Dict[str, Any] = {"type": "websocket.accept", "subprotocol": subprotocol}
        if headers and self._asgi_spec_version > 20:
            message["headers"] = self._encode_headers(headers)
        await self._send(message)
        self._accepted = True
        self.receive = self._wrapped_receive
        self.send = self._wrapped_send

    async def _wrapped_receive(self) -> Any:
        data = await self._receive()
        for method in self._flow_receive:  # type: ignore
            data = method(data)
        return data

    async def _wrapped_send(self, data: Any):
        for method in self._flow_send:  # type: ignore
            data = method(data)
        if isinstance(data, str):
            await self._send({"type": "websocket.send", "text": data})
        else:
            await self._send({"type": "websocket.send", "bytes": data})



from __future__ import annotations

import re
from abc import ABCMeta, abstractmethod
from http.cookies import SimpleCookie
from typing import Any, List, Mapping, Type, TypeVar, Union

from ...datastructures import sdict
from ...utils import cachedprop
from ..headers import Accept, LanguageAccept


AcceptType = TypeVar("AcceptType", bound=Accept)

_regex_accept = re.compile(
    r"([^\s;,]+(?:[ \t]*;[ \t]*(?:[^\s;,q][^\s;,]*|q[^\s;,=][^\s;,]*))*)" r"(?:[ \t]*;[ \t]*q=(\d*(?:\.\d+)?)[^,]*)?",
    re.VERBOSE,
)


class Wrapper:
    def __getitem__(self, name: str) -> Any:
        return getattr(self, name, None)

    def __setitem__(self, name: str, value: Any):
        setattr(self, name, value)


class IngressWrapper(Wrapper, metaclass=ABCMeta):
    __slots__ = ["scheme", "path"]

    scheme: str
    path: str

    @property
    @abstractmethod
    def headers(self) -> Mapping[str, str]: ...

    @cachedprop
    def host(self) -> str:
        return self.headers.get("host")

    def __parse_accept_header(self, value: str, cls: Type[AcceptType]) -> AcceptType:
        if not value:
            return cls(None)
        result = []
        for match in _regex_accept.finditer(value):
            mq = match.group(2)
            if not mq:
                quality = 1.0
            else:
                quality = max(min(float(mq), 1), 0)
            result.append((match.group(1), quality))
        return cls(result)

    @cachedprop
    def accept_language(self) -> LanguageAccept:
        return self.__parse_accept_header(self.headers.get("accept-language"), LanguageAccept)

    @cachedprop
    def cookies(self) -> SimpleCookie:
        cookies: SimpleCookie = SimpleCookie()
        for cookie in self.headers.get("cookie", "").split(";"):
            cookies.load(cookie)
        return cookies

    @property
    @abstractmethod
    def query_params(self) -> sdict[str, Union[str, List[str]]]: ...


class EgressWrapper(Wrapper, metaclass=ABCMeta):
    __slots__ = ["_proto"]

    def __init__(self, proto):
        self._proto = proto



import asyncio
from typing import AsyncGenerator

from ...http.response import HTTPBytesResponse, HTTPResponse
from ...http.wrappers.response import ResponseStream as _ResponseStream


class RequestCancelled(Exception): ...


class BodyWrapper:
    __slots__ = ["proto", "timeout", "_data"]

    def __init__(self, proto, timeout):
        self.proto = proto
        self.timeout = timeout
        self._data = bytearray()

    def _append_data(self, data: bytes):
        if data == b"":
            return
        self._data.extend(data)

    def __await__(self):
        if self.timeout:
            return self._await_with_timeout().__await__()
        return self._load_all().__await__()

    async def _await_with_timeout(self):
        try:
            rv = await asyncio.wait_for(self.proto(), timeout=self.timeout)
        except asyncio.TimeoutError:
            raise HTTPBytesResponse(408, b"Request timeout")
        return rv

    async def __aiter__(self) -> AsyncGenerator[bytes, None]:
        while True:
            event = await self.proto()
            if event["type"] == "http.request":
                yield event["body"]
                if not event.get("more_body", False):
                    break
            elif event["type"] == "http.disconnect":
                raise RequestCancelled

    async def _load_all(self):
        async for chunk in self:
            self._append_data(chunk)
        return bytes(self._data)


class ResponseStream(_ResponseStream):
    __slots__ = []

    async def __call__(self):
        for method in self.response._flow_stream:
            method()
        await self._proto(
            {
                "type": "http.response.start",
                "status": self.response.status,
                "headers": list(HTTPResponse.asgi_headers(self)),
            }
        )
        async for item in self._target:
            await self.send(self._item_wrapper(item))
        await self._proto({"type": "http.response.body", "body": b"", "more_body": False})
        return noop_response

    def send(self, data):
        if isinstance(data, str):
            data = data.encode("utf8")
        return self._proto({"type": "http.response.body", "body": data, "more_body": True})


class NoopResponse:
    async def asgi(self, scope, send):
        return


noop_response = NoopResponse()



from emmett_core.cryptography.symmetric import decrypt_b64, decrypt_hex, encrypt_b64, encrypt_hex


text = b"plain text"
key = "some key"


def test_b64():
    ct = encrypt_b64(text, key)
    assert decrypt_b64(ct, key) == text


def test_hex():
    ct = encrypt_hex(text, key)
    assert decrypt_hex(ct, key) == text



import pytest

from emmett_core.utils import _cached_prop_loop, _cached_prop_sync, cachedprop


class Class:
    def __init__(self):
        self.calls = 0

    @cachedprop
    def prop(self):
        self.calls += 1
        return "test_cachedprop_sync"

    @cachedprop
    async def prop_loop(self):
        self.calls += 1
        return "test_cachedprop_loop"


def test_cachedprop_sync():
    assert isinstance(Class.prop, _cached_prop_sync)
    obj = Class()
    assert obj.calls == 0
    assert obj.prop == "test_cachedprop_sync"
    assert obj.calls == 1
    assert obj.prop == "test_cachedprop_sync"
    assert obj.calls == 1


@pytest.mark.asyncio
async def test_cachedprop_loop():
    assert isinstance(Class.prop_loop, _cached_prop_loop)
    obj = Class()
    assert obj.calls == 0
    assert (await obj.prop_loop) == "test_cachedprop_loop"
    assert obj.calls == 1
    assert (await obj.prop_loop) == "test_cachedprop_loop"
    assert obj.calls == 1



import re

from ..datastructures import ImmutableList


class Accept(ImmutableList):
    def __init__(self, values=()):
        if values is None:
            list.__init__(self)
            self.provided = False
        elif isinstance(values, Accept):
            self.provided = values.provided
            list.__init__(self, values)
        else:
            self.provided = True
            values = sorted(values, key=lambda x: (x[1], x[0]), reverse=True)
            list.__init__(self, values)

    def _value_matches(self, value, item):
        return item == "*" or item.lower() == value.lower()

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.quality(key)
        return list.__getitem__(self, key)

    def quality(self, key):
        for item, quality in self:
            if self._value_matches(key, item):
                return quality
        return 0

    def __contains__(self, value):
        for item, _quality in self:
            if self._value_matches(value, item):
                return True
        return False

    def __repr__(self):
        return "%s([%s])" % (self.__class__.__name__, ", ".join("(%r, %s)" % (x, y) for x, y in self))

    def index(self, key):
        if isinstance(key, str):
            for idx, (item, _quality) in enumerate(self):
                if self._value_matches(key, item):
                    return idx
            raise ValueError(key)
        return list.index(self, key)

    def find(self, key):
        try:
            return self.index(key)
        except ValueError:
            return -1

    def values(self):
        for item in self:
            yield item[0]

    def to_header(self):
        result = []
        for value, quality in self:
            if quality != 1:
                value = "%s;q=%s" % (value, quality)
            result.append(value)
        return ",".join(result)

    def __str__(self):
        return self.to_header()

    def best_match(self, matches, default=None):
        best_quality = -1
        result = default
        for server_item in matches:
            for client_item, quality in self:
                if quality <= best_quality:
                    break
                if self._value_matches(server_item, client_item) and quality > 0:
                    best_quality = quality
                    result = server_item
        return result

    @property
    def best(self):
        if self:
            return self[0][0]


class LanguageAccept(Accept):
    regex_locale_delim = re.compile(r"[_-]")

    def _value_matches(self, value, item):
        def _normalize(language):
            return self.regex_locale_delim.split(language.lower())[0]

        return item == "*" or _normalize(value) == _normalize(item)
